from random import randint

from .base import BaseMethodsGroup

class Messages(BaseMethodsGroup):

    async def getConversationMembers(self,
                                     peer_id: int,
                                     offset: int | None = None,
                                     count: int | None = None,
                                     extended: bool = False,
                                     fields: str | None = None,
                                     group_id: int | None = None):
        '''
        Метод получает список участников беседы.
        (см. https://dev.vk.ru/ru/method/messages.getConversationMembers)
        '''
        values = {
            "peer_id": peer_id,
            "offset": abs(offset) if offset is not None else None,
            "count": abs(count) if count is not None else None,
            "extended": "true" if extended else "false",
            "fields": fields,
            "group_id": abs(group_id) if group_id is not None else self.__group_id__,
            "v": self.__api_version__
        }
        response = await self.base_api.base_get_method(api_method=f"{self.method}.getConversationMembers",
                                                       values=values)
        return response

    async def getConversationById(self,
                                  peer_ids: int | list[int],
                                  extended: bool = False,
                                  fields: str | None = None,
                                  group_id: int | None = None):
        '''
        Позволяет получить беседу по её идентификатору.
        (см. https://dev.vk.ru/ru/method/messages.getConversationsById)
        '''
        values = {
            "peer_ids": ",".join([f"{pid}" for pid in peer_ids]) if isinstance(peer_ids, list) else f"{peer_ids}",
            "extended": "true" if extended else "false",
            "fields": fields,
            "group_id": abs(group_id) if group_id is not None else self.__group_id__,
            "v": self.__api_version__
        }
        response = await self.base_api.base_get_method(api_method=f"{self.method}.getConversationsById",
                                                       values=values)
        return response
    
    async def removeChatUser(self,
                             chat_id: int,
                             user_id: int | None = None,
                             member_id: int | None = None):
        '''
        Исключает из мультидиалога пользователя, если текущий пользователь или сообщество
        является администратором беседы либо текущий пользователь пригласил исключаемого пользователя.
        (см. https://dev.vk.ru/ru/method/messages.removeChatUser)
        '''
        values = {
            "chat_id": abs(chat_id),
            "user_id": user_id,
            "member_id": member_id,
            "v": self.__api_version__
        }
        response = await self.base_api.base_get_method(api_method=f"{self.method}.removeChatUser",
                                                       values=values)
        return response
    
    async def send(self,
                   user_id: int = None,
                   peer_id: int = None,
                   peer_ids: int | list[int] = None,
                   domain: str = None,
                   chat_id: int = None,
                   user_ids: int | list[int] = None,
                   message: str = None,
                   lat: str = None,
                   long: str = None,
                   attachment: str | list[str] = None,
                   reply_to: int = None,
                   forward_messages: int | list[int] = None,
                   forward: dict = None,
                   sticker_id: int = None,
                   group_id: int | None = None,
                   keyboard: object = None,
                   template: object = None,
                   payload: object = None,
                   content_source: dict = None,
                   dont_parse_links: bool = None,
                   disable_mentions: bool = None,
                   intent: str = None,
                   subsribe_id: int = None):
        
        '''
        Отправляет сообщение
        (см. https://dev.vk.ru/ru/method/messages.send)
        '''

        values = {
            "user_id": user_id,
            "random_id": randint(0, 1000000000),
            "peer_id": peer_id,
            "peer_ids": ",".join([f"{pid}" for pid in peer_ids]) if isinstance(peer_ids, list) else f"{peer_ids}",
            "domain": domain,
            "chat_id": chat_id,
            "user_ids": ",".join([f"{uid}" for uid in user_ids]) if isinstance(user_ids, list) else f"{user_ids}",
            "message": message,
            "lat": lat,
            "long": long,
            "attachment": ",".join([f"{attach}" for attach in attachment]) if attachment is not None else None,
            "reply_to": reply_to,
            "forward_messages": forward_messages,
            "forward": forward,
            "sticker_id": abs(sticker_id) if sticker_id is not None else None,
            "group_id": abs(group_id) if group_id is not None else self.__group_id__,
            "keyboard": keyboard,
            "template": template,
            "payload": payload,
            "content_source": content_source,
            "dont_parse_links": "true" if dont_parse_links else "false",
            "disable_mentions": "true" if disable_mentions else "false",
            "intent": intent,
            "subscribe_id": abs(subsribe_id) if subsribe_id is not None else None,
            "v": self.__api_version__
        }
        response = await self.base_api.base_get_method(api_method=f"{self.method}.send",
                                                       values=values)
        return response


import asyncio

from ..api.twilight_api import TwilightAPI
from .api import FrameworkRouter
from .polling.bots_longpoll import BotsLongPoll
from ..logger.darky_logger import DarkyLogger
from ..logger.darky_visual import STYLE, FG
from ..components.logo import LogoComponent
from ..utils.config import CONFIG
from ..utils.types.twi_states import TwiVKStates
from ..utils.event_loop import TwiTaskManager
from .exceptions.framework import (
    InitializationError
)

class TwilightVK:

    def __init__(self,
                 BOT_NAME: str = None,
                 ACCESS_TOKEN: str = None, 
                 GROUP_ID: int = None, 
                 API_VERSION: str = CONFIG.VK_API.version,
                 API_MODE: str = "BOTSLONGPOLL",
                 TWI_API_ENABLED: bool = True,
                 HOST: str = CONFIG.API.host,
                 PORT: str = CONFIG.API.port,
                 loop_wrapper: TwiTaskManager = None) -> None:
        '''
        Initializes TwilightVK

        :param BOT_NAME: name of your bot
        :type BOT_NAME: str

        :param ACCESS_TOKEN: your group's access token, you can find it here(https://dev.vk.com/ru/api/access-token/community-token/in-community-settings)
        :type ACCESS_TOKEN: str

        :param GROUP_ID: your group's id, you can find it in your group's settings
        :type GROUP_ID: int

        :param API_VERSION: version of VK API, by default its grabbed from the frameworks's default configuration
        :type API_VERSION: str | None

        :param API_MODE: mode of polling (BOTSLONGPOLL/.../...)
        :type API_MODE: str

        :param TWI_API_ENABLED: should be bot's api awailable
        :type TWI_API_ENABLED: bool

        :param loop_wrapper: Initialized class TwiTaskManager for async loop wrapping
        :type loop_wrapper: TwiTaskManager
        '''
        self._state = TwiVKStates.INITIALIZING
        self.bot_name = self.__class__.__name__ if BOT_NAME is None else BOT_NAME
        self.logo = LogoComponent()
        self.logger = DarkyLogger(logger_name=f"twilight-vk", configuration=CONFIG.LOGGER)

        self.logger.info(f"Initializing framework...")

        try:
            if not ACCESS_TOKEN or not GROUP_ID:
                raise InitializationError(ACCESS_TOKEN, GROUP_ID)
        except InitializationError as ex:
            self.logger.critical(f"Initialization error{ex}")
            exit()

        self.__access_token__ = ACCESS_TOKEN
        self.__group_id__ = GROUP_ID
        self.__api_version__ = API_VERSION

        self._loop_wrapper = TwiTaskManager() if loop_wrapper is None else loop_wrapper

        API_MODES = {
            "BOTSLONGPOLL": BotsLongPoll(access_token=ACCESS_TOKEN,
                                         group_id=GROUP_ID,
                                         api_version=API_VERSION,
                                         loop_wrapper=self._loop_wrapper)
        }
        self.__bot__ = API_MODES.get(API_MODE, "BOTSLONGPOLL")
        self.methods = self.__bot__.vk_methods
        self.on_event = self.__bot__._router.on_event

        self.api_router = FrameworkRouter(self)

        self._api: TwilightAPI = None

        #if TWI_API_ENABLED:
        if False:
            self._api = TwilightAPI(
                BOTS = [self],
                HOST = HOST,
                PORT = PORT,
                loop_wrapper = self._loop_wrapper,
                _need_root_router = False
            )
            self._loop_wrapper.add_task(self._api.run_server())

        self._state = TwiVKStates.DISABLED
        self.logger.info(f"Framework initialized")

    async def run_polling(self):
        '''
        Start polling
        '''
        try:
            self._state = TwiVKStates.STARTING
            self.logger.info(f"Framework is starting...")

            await self.__bot__.auth()

            if self.__bot__.__server__ is not None:
                self._state = TwiVKStates.READY
                self.logger.info(f"{FG.GREEN}Framework is started (BOT_NAME: {self.bot_name}){STYLE.RESET}")
            else:
                self.logger.error(f"Server was not aquired. Exiting...")
                self.should_stop()

            async for event_response in self.__bot__.listen():
                if self._state == TwiVKStates.READY:
                    self._loop_wrapper.add_task(self.__bot__._router.handle(event_response))

        except KeyboardInterrupt:
            self.logger.debug(f"TwilightVK was stopped by KeyboardInterrupt")
        except asyncio.CancelledError:
            self.logger.warning(f"Polling was forcibly canceled (it is not recommend to do this)")
        except Exception as exc:
            self.logger.critical(f"Framework was crashed with critical unhandled error", exc_info=True)
            self._state = TwiVKStates.ERROR
        finally:
            #self.__loop__.stop()
            if self._state == TwiVKStates.READY and not self.__bot__.__stop__:
                self.__bot__.stop()
            self.logger.info(f"{FG.RED}Framework has been stopped{STYLE.RESET}")
            await asyncio.sleep(0.1)
            if self._state not in [TwiVKStates.ERROR]:
                self._state = TwiVKStates.DISABLED
            if self._api:
                await self._api.stop()


    def start(self):
        '''
        Starts the bot and polling until stop() is called
        '''
        try:
            self.logger.note(self.logo.colored)
            self._loop_wrapper.add_task(self.run_polling())
            self._loop_wrapper.run()
        except KeyboardInterrupt:
            self.logger.debug(f"TwilightVK was stopped by KeyboardInterrupt")
        except Exception:
            self.logger.critical("Unhandled error", exc_info=True)

    def stop(self, force: bool = False):
        '''
        Stops the polling and bot

        :param force: *[Optional]* Force stopping with cancelling all current event handlings
        :type force: bool
        '''
        if self._state == TwiVKStates.READY:

            self.logger.info(f"Shutting down...")
            self.should_stop()

            #if force:
            #    self.logger.warning(f"Forced stop. For soft stop - use TwilightVK.should_stop() method")
            #    self._loop_wrapper.cancel_tasks(targets = [self.run_polling()])
    
    def should_stop(self):
        '''
        Tells the bot that it should stop after the next event
        '''
        if self._state == TwiVKStates.READY and not self.__bot__.__stop__:
            self.logger.info("Framework was asked to stop")
            self._state = TwiVKStates.SHUTTING_DOWN
            self.__bot__.stop()
        elif self._state == TwiVKStates.SHUTTING_DOWN:
            self.logger.warning("Framework is already stopping")
        else:
            self.logger.error(f"Unable to stop framework for some reason. BOT_STATE={self._state}")

    def __getApiRouters__(self):
        return self.api_router.get_router()


import asyncio
from pathlib import Path
from typing import TYPE_CHECKING

import uvicorn
from fastapi import FastAPI, APIRouter
from fastapi.responses import FileResponse
from fastapi.openapi.docs import get_swagger_ui_html
from contextlib import asynccontextmanager

from ..components.logo import LogoComponent
from ..logger.darky_logger import DarkyLogger
from ..logger.darky_visual import STYLE, FG, BG, Visual
from ..utils.config import CONFIG
from ..utils.event_loop import TwiTaskManager
from ..utils.types.twi_states import TwiAPIStates
from .routers import RootRouter

BASE_DIR = Path(__file__).resolve().parent
ASSETS = BASE_DIR / "assets"

Visual.ansi()

class TwilightAPI:

    def __init__(self,
                 BOTS: object | list[object] = [],
                 HOST: str = CONFIG.API.host,
                 PORT: str = CONFIG.API.port,
                 loop_wrapper: TwiTaskManager = None,
                 _need_root_router: bool = True
                ):
        '''
        API Swagger for bots based on Twilight framework

        :param BOTS: One or list of initialized Twilight bot objects
        :type BOTS: object | list[object]

        :param HOST: Sets the host ip adress for API Swagger
        :type HOST: str

        :param PORT: Sets the port for API Swagger
        :type PORT: str
        '''
        self._state = TwiAPIStates.INITIALIZING
        self.logo = LogoComponent()
        self.logger = DarkyLogger(logger_name=f"twilight-api", configuration=CONFIG.LOGGER)
        
        self.logger.info(f"Initializing API...")

        self.__HOST__ = HOST
        self.__PORT__ = PORT

        self.__api__ = FastAPI(
            title=CONFIG.API.title,
            description=CONFIG.API.description,
            version=CONFIG.API.version,
            lifespan=self.lifespan,
            root_path=CONFIG.API.prefix,
            docs_url=None
        )

        self.bots:list = BOTS

        self._loop_wrapper = TwiTaskManager() if loop_wrapper is None else loop_wrapper

        self._uvicorn_config = uvicorn.Config(
                app=self.__api__,
                host=self.__HOST__,
                port=self.__PORT__,
                log_config=CONFIG.LOGGER,
            )
        
        if _need_root_router:
            self.__api__.include_router(RootRouter(self.bots, self).get_router())

        self.router = APIRouter(
            tags=["Server"]
        )
        self.router.add_api_route(path="/favicon.ico", endpoint=self.favicon, methods=["GET"],
                                  include_in_schema=False)
        self.router.add_api_route(path="/docs", endpoint=self.custom_swagger, methods=["GET"],
                                  include_in_schema=False)
        self.__api__.include_router(self.router)

        self.logger.debug(f"Importing bot's API routers...")
        for bot in self.bots:
            self.logger.debug(f"Importing API routers from {bot.__class__.__name__}<{self.bots.index(bot)}> - {bot.bot_name}...")
            self.__api__.include_router(bot.__getApiRouters__())
        
        if not self.bots:
            self.logger.warning(f"There is no connected bots to the API")
        
        self._state = TwiAPIStates.DISABLED
        self.logger.info(f"API initialized")

    @asynccontextmanager
    async def lifespan(self, api: FastAPI):
        '''
        Provides start_up(preparing API wrapper) and
        shut_down(saving data for example) event handling
        '''
        try:
            '''Here is the startup code'''
            self._state = TwiAPIStates.STARTING
            self.logger.info(f"Twilight API is starting...")
            ...
            self._state = TwiAPIStates.READY
            self.logger.info(f"{FG.GREEN}Twilight API is started (on {CONFIG.API.host}:{CONFIG.API.port}){STYLE.RESET}")

            yield

            '''Here is the shutdown code'''
            self._state = TwiAPIStates.SHUTTING_DOWN
            ...
            self._state = TwiAPIStates.DISABLED
            self.logger.info(f"{FG.RED}Twilight API is stopped{STYLE.RESET}")
        except Exception as ex:
            self.logger.critical(f"Unhandled error", exc_info=True)
            await self.stop()

    async def run_server(self):
        '''
        Runs the API server
        '''
        try:
            self.__uvicorn_server__ = uvicorn.Server(self._uvicorn_config)
            await self.__uvicorn_server__.serve()
        except KeyboardInterrupt:
            self.logger.debug("API was stopped by KeyboardInterrupt")

    def start(self):
        '''
        Starts the API Swagger
        '''
        try:
            self.logger.note(self.logo.colored)
            self.__uvicorn_server__ = uvicorn.Server(self._uvicorn_config)
            self._loop_wrapper.add_task(self.run_server())
            self._loop_wrapper.run()
        except KeyboardInterrupt:
            self.logger.debug("API was stopped by KeyboardInterrupt")
        except Exception:
            self.logger.critical("Unhandled error", exc_info=True)

    async def stop(self):
        '''
        Shutdown the API Swagger
        '''
        if self.__uvicorn_server__.started:
            self.logger.info(f"Gracefully shutting down the API...")
            self.__uvicorn_server__.should_exit = True

            try:
                await asyncio.wait_for(
                    asyncio.sleep(1.0),
                    timeout = 5.0
                )
            except asyncio.TimeoutError:
                self.logger.warning("Force stopping API after timeout")
            
            self.logger.debug("Shutting down complete")
    
    #---
    #API METHODS
    #---

    async def favicon(self) -> FileResponse:
        '''
        THIS IS API METHOD
        '''
        return FileResponse(ASSETS / "favicon.ico")
    
    async def custom_swagger(self):
        '''
        THIS IS API METHOD
        '''
        return get_swagger_ui_html(
            openapi_url=self.__api__.openapi_url,
            title = f"{self.__api__.title} - Swagger UI",
            swagger_favicon_url="/favicon.ico"
        )
    
    #TODO: importing routers from bots


from ..logger.formatters import (
    DarkyConsoleFormatter,
    DarkyFileFormatter,
    UvicornAccessFormatter
)

class CONFIG:

    class FRAMEWORK:
        version = "0.1.0-beta1"
        developer = "darky_wings"
    
    class VK_API:
        url = "https://api.vk.ru"
        version = "5.199"
        wait = 25

    class API:
        host = "0.0.0.0"
        port = 8000
        title = "Twilight API Swagger"
        description = "Welcome to the Twilight API Swagger!"
        version = "0.0.1"
        prefix = "/api/v1"

    LOGGER = {
                "version": 1,
                "disable_existing_loggers": False,
                "formatters": {
                    "file": {
                        "()": DarkyFileFormatter,
                        "fmt": "%(name)s | %(asctime)s | %(levelname)s | %(message)s"
                    },
                    "console": {
                        "()": DarkyConsoleFormatter,
                        "fmt": "%(name)s | %(asctime)s | %(levelname)s | %(message)s",
                        "colored": True,
                        "color_core_name": True
                    },
                    "uvicorn_access_console": {
                        "()": UvicornAccessFormatter,
                        "fmt": "%(name)s | %(asctime)s | %(levelname)s | %(client_addr)s - \"%(request_line)s\" %(status_code)s",
                        "colored": True
                    }
                },
                "handlers": {
                    "file": {
                        "level": "INIT",
                        "class": "logging.handlers.RotatingFileHandler",
                        "formatter": "file",
                        "filename": "twilight_vk.log",
                        "backupCount": 3,
                        "encoding": "utf-8"
                    },
                    "console": {
                        "level": "INFO",
                        "class": "logging.StreamHandler",
                        "formatter": "console"
                    },
                    "uvicorn_access_console": {
                        "level": "INIT",
                        "class": "logging.StreamHandler",
                        "formatter": "uvicorn_access_console"
                    }
                },
                "loggers": {
                    "twilight-api": {
                        "handlers": ["console", "file"],
                        "level": "INIT",
                        "propagate": True
                    },
                    "uvicorn.access": {
                        "handlers": ["uvicorn_access_console", "file"],
                        "level": "INIT",
                        "propagate": True
                    },
                    "uvicorn.error": {
                        "handlers": ["uvicorn_access_console", "file"],
                        "level": "WARNING",
                        "propagate": True
                    },
                    "twi-api-fw": {
                        "handlers": ["console", "file"],
                        "level": "INIT",
                        "propagate": True
                    },
                    "twi-api-vkapi": {
                        "handlers": ["console", "file"],
                        "level": "INIT",
                        "propagate": True
                    },
                    "twilight-vk": {
                        "handlers": ["console", "file"],
                        "level": "INIT",
                        "propagate": True
                    },
                    "botslongpoll": {
                        "handlers": ["console", "file"],
                        "level": "INIT",
                        "propagate": True
                    },
                    "vk-methods": {
                        "handlers": ["console", "file"],
                        "level": "INIT",
                        "propagate": True
                    },
                    "http-validator": {
                        "handlers": ["console", "file"],
                        "level": "INIT",
                        "propagate": True
                    },
                    "event-validator": {
                        "handlers": ["console", "file"],
                        "level": "INIT",
                        "propagate": True
                    },
                    "event-router": {
                        "handlers": ["console", "file"],
                        "level": "INIT",
                        "propagate": True
                    },
                    "event-handler": {
                        "handlers": ["console", "file"],
                        "level": "INIT",
                        "propagate": True
                    },
                    "rule-handler": {
                        "handlers": ["console", "file"],
                        "level": "INIT",
                        "propagate": True
                    },
                    "loop-manager": {
                        "handlers": ["console", "file"],
                        "level": "INIT",
                        "propagate": True
                    },
                    "asyncio": {
                        "handlers": ["console", "file"],
                        "level": "INFO",
                        "propagate": True
                    }
                }
            }


import logging

from aiohttp import ClientResponse

from ...utils.config import CONFIG
from ..exceptions.validator import (
    EventValidationError
)
from ..exceptions.vkapi import (
    VkApiError,
    AuthError,
    LongPollError
)

class EventValidator:

    def __init__(self):
        self.__requiredFields__ = ['response', 'error']
        self.__pollingRequiredFileds__ = ['updates', 'failed']
        self.__errorFields__ = [
            self.__requiredFields__[-1],
            self.__pollingRequiredFileds__[-1]
            ]
        self.logger = logging.getLogger("event-validator")

    async def __isJsonValid__(self,
                              response:ClientResponse) -> bool:
        self.logger.debug(f"Checking Content-Type header for JSON...")
        if 'application/json' not in response.headers.get('Content-Type', '').lower():
            self.logger.error(f"Response doesn't have JSON content")
            return False
        return True
    
    async def __fieldsAreValid__(self,
                                 content:dict,
                                 fields:dict={}) -> bool:
        self.logger.debug(f"Checking for one of the required response fields...")
        for field in fields:
            if field in content:
                return True
        else:
            self.logger.error(f"Response doesn't contain any of the requirement fields")
            return False
    
    async def __haveErrors__(self,
                             content:dict) -> bool:
        self.logger.debug(f"Checking for errors in response...")
        for field in self.__errorFields__:
            if field in content:
                # TODO: VK API Exceptions and validators
                if field == "error":
                    error_code = content[field]["error_code"]
                    error_msg = content[field]["error_msg"]
                    request_params = content[field]["request_params"]
                    
                    if content[field]["error_code"] in [5, 1116]:
                        raise AuthError(error_code, error_msg, request_params)
                    
                    raise VkApiError(error_code, error_msg, request_params)
                if field == "failed":
                    failed_code = content[field]
                    raise LongPollError(failed_code, content["ts"] if "ts" in content else None)
                return True
        else:
            return False

    async def validate(self,
                 response:ClientResponse,
                 from_polling:bool=False) -> dict:
        self.logger.debug(f"Validating event response for [{response.method}] \"{response.url}\"...")

        jsonIsValid = await self.__isJsonValid__(response)

        if jsonIsValid:
            content = await response.json()
            fieldsAreValid = await self.__fieldsAreValid__(content, self.__requiredFields__ if not from_polling else self.__pollingRequiredFileds__)
            haveErrors = await self.__haveErrors__(content)

            if fieldsAreValid and not haveErrors:
                self.logger.debug("Event response is valid")
                return content
        
            self.logger.warning(f"Event response is not valid")
            self.logger.warning(f'{"jsonIsValid":<15}|{"":>1}{"fieldsAreValid":<15}|{"":>1}{"haveErrors":<15}|{"":>1}{"content":<15}')
            self.logger.warning(f'{jsonIsValid:<15}|{"":>1}{fieldsAreValid:<15}|{"":>1}{haveErrors:<15}|{"":>1}{content}')

        raise EventValidationError(jsonIsValid, fieldsAreValid, content)


import pytest

import asyncio
from aiohttp import ClientResponse

from twilight_vk import *
from twilight_vk.http.async_http import Http

@pytest.fixture
def bot():
    event_loop = asyncio.new_event_loop()
    asyncio.set_event_loop(event_loop)
    testBot = TwilightVK(
        ACCESS_TOKEN="123",
        GROUP_ID=123
    )
    yield testBot

class MockGetLongPollServer():

    def __init__(self, isSuccess=True):
        self.method = "GET"
        self.url = "https://fakeauth"
        self.headers = {"Content-Type": "application/json"}
        self.is_success = isSuccess

    async def json(self):
        if self.is_success:
            return {
                "response": {
                                "server": "https://fakeserverurl",
                                "key": "fakekey",
                                "ts": 123
                            }
            }
        return {
                    "error": {
                        "error_code": 5,
                        "error_msg": "User authorization failed: invalid access_token (4).",
                        "request_params": [
                            {
                                "key": "access_token",
                                "value": "123"
                            },
                            {
                                "key": "group_id",
                                "value": 123
                            },
                            {
                                "key": "v",
                                "value": "1.234"
                            }
                        ]
                    }
            }

class MockPollingEvent():

    def __init__(self,
                 event_type: str):
        self.method = "GET"
        self.url = "https://fakeapiserver"
        self.headers = {"Content-Type": "application/json"}
        self.status = 200
        self.event_type = event_type

    async def json(self):
        match self.event_type:
            case "message_typing_state":
                return {
                    "ts": 123,
                    "updates": [
                        {
                            "group_id": 123,
                            "type": "message_typing_state",
                            "event_id": "abc123",
                            "v": "1.234",
                            "object": {
                                "from_id": 123,
                                "to_id": -123,
                                "state": "typing"
                            }
                        }
                    ]
                }
            case "message_new":
                return {
                    "ts": 123,
                    "updates": [
                        {
                            "group_id": 123,
                            "type": "message_new",
                            "event_id": "abc123",
                            "v": "1.234",
                            "object": {
                                "client_info": {},
                                "message": {
                                    "date": 1000000000,
                                    "from_id": 123,
                                    "id": 1,
                                    "fwd_messages": [],
                                    "attachments": [],
                                    "conversation_message_id": 1,
                                    "text": "a",
                                    "peer_id": 2000000001,
                                    "random_id": 0
                                }
                            }
                        }
                    ]
                }
            case "failed 1":
                return {"failed": 1,
                        "ts": 30}
            case "failed 2":
                return {"failed": 2}
            case "failed 3":
                return {"failed": 3}

async def mock_polling(bot: TwilightVK, type: str, monkeypatch):
    async def fake_GetHttpEvent(
            url: str,
            params: dict = {},
            headers: dict = {},
            raw: bool = True
    ):
        fake_response = MockPollingEvent(type)
        return fake_response
    
    async def fake_typeMatch(response: ClientResponse,
                             self = bot.__bot__.httpValidator):
        return True

    monkeypatch.setattr(bot.__bot__.httpValidator, "__isValid__", fake_typeMatch)
    monkeypatch.setattr(bot.__bot__.httpClient, "get", fake_GetHttpEvent)

    return bot

@pytest.mark.asyncio
async def test_authorization(bot: TwilightVK, caplog, monkeypatch):

    async def fake_getBotLongPollServer():
        fake_response = MockGetLongPollServer(False)
        return await bot.__bot__.eventValidator.validate(fake_response)
    
    monkeypatch.setattr(bot.__bot__.vk_methods.groups, "getLongPollServer", fake_getBotLongPollServer)

    await bot.__bot__.auth()
    assert bot.__bot__.__server__ is None
    assert bot.__bot__.__key__ is None
    assert bot.__bot__.__ts__ is None
    assert "Authrization error: [5] User authorization failed: invalid access_token (4)." in caplog.text

    async def fake_getBotLongPollServer():
        fake_response = MockGetLongPollServer(True)
        return await bot.__bot__.eventValidator.validate(fake_response)
    
    monkeypatch.setattr(bot.__bot__.vk_methods.groups, "getLongPollServer", fake_getBotLongPollServer)

    await bot.__bot__.auth()
    assert bot.__bot__.__server__ is not None
    assert bot.__bot__.__key__ is not None
    assert bot.__bot__.__ts__ is not None
    assert "Authorized" in caplog.text

@pytest.mark.asyncio
async def test_polling_check_event(bot: TwilightVK, caplog, monkeypatch):
    bot = await mock_polling(bot, "message_typing_state", monkeypatch)
    assert await bot.__bot__.check_event() == await MockPollingEvent("message_typing_state").json()
    
@pytest.mark.asyncio
async def test_polling_listen(bot: TwilightVK, caplog, monkeypatch):
    bot = await mock_polling(bot, "message_new", monkeypatch)
    async for event in bot.__bot__.listen():
        assert event == await MockPollingEvent("message_new").json()
        break

@pytest.mark.asyncio
async def test_polling_failed(bot: TwilightVK, caplog, monkeypatch):

    bot._state = "ready"

    bot.should_stop()

    async def fake_getBotLongPollServer():
        fake_response = MockGetLongPollServer(True)
        return await bot.__bot__.eventValidator.validate(fake_response)
    
    monkeypatch.setattr(bot.__bot__.vk_methods.groups, "getLongPollServer", fake_getBotLongPollServer)

    failed_codes = ["1", "2", "3"]

    for code in failed_codes:
        bot = await mock_polling(bot, f"failed {code}", monkeypatch)

        await bot.__bot__.check_event()

        if code == "1":
            assert bot.__bot__.__ts__ == 30
        
    assert "The event history is outdated or has been partially lost. The application can receive events further using the new \"ts\" value from the response." in caplog.text
    assert "The key is expired. Getting new one..." in caplog.text
    assert "The information is lost. Reauthorizing..." in caplog.text

@pytest.mark.asyncio
async def test_polling_stop(bot: TwilightVK, caplog, monkeypatch):

    bot._state = "ready"
    
    bot.should_stop()
    assert "Polling will be stopped as soon as the current request will be done. Please wait" in caplog.text


from .base import BaseRule
from ...utils.twiml import TwiML
from ...utils.types.event_types import MessageActionTypes

class TrueRule(BaseRule):

    '''
    Возвращает всегда True
    Правило сделано в основном для теста
    '''
    async def check(self, event: dict):
        return True

class FalseRule(BaseRule):

    '''
    Возвращает всегда False
    Правило сделано в основном для теста
    '''
    async def check(self, event: dict):
        return False

class ContainsRule(BaseRule):

    def __init__(self,
                 triggers: str | list[str],
                 ignore_case: bool = False,
                 need_list: bool = True):
        '''
        Проверяет наличие указанных в value фрагментов текста в сообщении, возвращает True при нахождении

        :param triggers: Слова-триггеры
        :type triggers: str | list[str]

        :param ignore_case: Флаг игнорирования регистра
        :type ignore_case: bool

        :param need_list: Дает понять нужно ли возвращать список найденных фрагментов или достаточно просто оповестить
        :type need_list: bool
        '''
        super().__init__(
            triggers = triggers,
            ignore_case = ignore_case,
            need_list = need_list
        )
    
    async def check(self, event: dict) -> bool | dict:
        text: str = event["object"]["message"]["text"]
        text = text.lower() if self.ignore_case else text
        result = {
            "triggers": []
        }

        for trigger in self.triggers:
            if trigger in text:
                if self.need_list:
                    result["triggers"].append(trigger)
                    continue
                return True
            
        if result["triggers"] != []:
            return result
        
        return False


class TextRule(BaseRule):

    def __init__(self,
                 value: str | list[str],
                 ignore_case: bool = False):
        '''
        Сверяет текст сообщения с указанным, возвращает True при совпадении

        :param value: Ожидаемое значение
        :type value: str | list[str]

        :param ignore_case: Флаг игнорирования регистра
        :type ignore_case: bool
        '''

        super().__init__(
            value = value,
            ignore_case = ignore_case
        )

    async def check(self, event: dict) -> bool:
        text:str = event["object"]["message"]["text"]
        return (text.lower() if self.ignore_case else text) in \
            ([val.lower() if self.ignore_case else self.value for val in self.value])


class TwiMLRule(BaseRule):

    def __init__(self,
                 value:str|list[str],
                 ignore_case:bool=False):
        '''
        Сверяет текст сообщения с указанным шаблоном на основе regex выражений, 
        возвращает словарь найденных аргументов или False если сообщение не соответствует шаблону

        :param value: Ожидаемое выражение(шаблон)
        :type value: str | list[str]

        :param ignore_case: Флаг игнорирования регистра
        :type ignore_case: bool
        '''

        super().__init__(
            value = value,
            ignore_case = ignore_case
        )

    async def check(self, event: dict) -> bool | dict:
        text:str = event["object"]["message"]["text"]
        text = text.lower() if self.ignore_case else text

        twiml = TwiML()

        for value in self.value:
            value = value.lower() if self.ignore_case else value

            twiml.update_template(value)
            result = await twiml.parse(text)

            if result is not None:
                return result
        else:
            return False


class MentionRule(BaseRule):

    def __init__(self,
                 need_list: bool = True):
        '''
        Проверяет наличие упоминаний в сообщении
        возвращает словарь найденных упоминаний/True или False если ни одного упоминания не было в сообщении
        
        :param need_list: Дает понять нужно ли возвращать список упоминаний или достаточно просто оповестить что упоминание было
        :type need_list: bool
        '''
        super().__init__(
            need_list = need_list
        )
    
    async def __get_mentions__(self, event: dict) -> dict:
        text: str = event["object"]["message"]["text"]

        twiml = TwiML()
        result = await twiml.extract_mentions(text)

        return result

    async def check(self, event: dict) -> bool | dict:
        result = await self.__get_mentions__(event)

        if result == {"mentions": []}:
            return False
        
        if self.need_list:
            return result
        
        return True

class IsMentionedRule(MentionRule):
    
    '''
    Проверяет был ли упомянут сам бот или нет
    '''
    async def check(self, event: dict) -> bool:
        mentions = await self.__get_mentions__(event)

        if mentions != False:
            for mention in mentions["mentions"]:
                if mention["type"] == "club" and mention["id"] == event.get("group_id", 0):
                    return True
        
        return False


class ReplyRule(BaseRule):

    '''
    Проверяет наличие ответа в событии.
    Возвращает True/False в зависимости от результата
    '''
    async def check(self, event: dict) -> bool:
        if event["object"]["message"].get("reply_message", None) is not None:
            return True
        return False

class ForwardRule(BaseRule):

    '''
    Проверяет наличие пересланного сообщения в событии.
    Возвращает True/False в зависимости от результата.
    '''
    async def check(self, event: dict) -> bool:
        if event["object"]["message"].get("fwd_messages") != []:
            return True
        return False


class AdminRule(BaseRule):

    '''
    Проверяет является ли пользователь отправивший сообщение в беседе его администратором
    Возвращает True/False в зависимости от результата
    '''

    async def __get_admins__(self, event: dict) -> None:
        if (
            event.setdefault("is_admin", None) is None or
            event.setdefault("is_bot_admin", None) is None
        ):
            chat_members = await self.methods.messages.getConversationMembers(
                peer_id=event["object"]["message"]["peer_id"]
            )
            member: dict
            for member in chat_members["response"]["items"]:
                if member.get("member_id") == event["object"]["message"]["from_id"]:
                    event["is_admin"] = member.get("is_admin", False)
                if member.get("member_id") == -event.get("group_id"):
                    event["is_bot_admin"] = member.get("is_admin", False)

    async def check(self, event: dict) -> bool:
        await self.__get_admins__(event)
        return event.get("is_admin")

class IsAdminRule(AdminRule):

    '''
    Проверяет является ли бот администратором в чате
    Возвращает True/False в зависимости от результата
    '''
    async def check(self, event: dict) -> bool:
        await self.__get_admins__(event)
        return event.get("is_bot_admin")


class InvitedRule(BaseRule):
    
    '''
    Возвращает True если какой-либо пользователь был добавлен в чат.
    '''
    async def __who_is_invited__(self, event: dict) -> int:
        if (
            event["object"]["message"].get("action", None) is not None and
            event["object"]["message"]["action"]["type"] == MessageActionTypes.CHAT_INVITE_USER
        ):
            return event["object"]["message"]["action"]["member_id"]
        return 0

    async def check(self, event: dict) -> bool:
        if await self.__who_is_invited__(event) != 0:
            return True
        return False 
    
class IsInvitedRule(InvitedRule):

    '''
    Возвращает True если бот был добавлен в чат
    '''
    async def check(self, event: dict) -> bool:
        if await self.__who_is_invited__(event) == -event.get("group_id"):
            return True
        return False


import pytest

from twilight_vk.utils.twiml import TwiML

@pytest.fixture
def twiml():
    twimlInstance = TwiML()
    yield twimlInstance

@pytest.fixture
def parse_dataset():
    return {
        "messages": [
            "",
            "test",
            "test 123",
            "test darky",
            "test darky 21",
            "test darky darky",
            "test darky abc efg"
        ],
        "templates": [
            "test",
            "test <user>",
            "test <user:int>",
            "test <user:word>",
            "test <user:word> <age:int>",
            "test <user> <testword:word>"
        ],
        "results": [
            [None, {}, {}, {}, {}, {}, {}],
            [None, None, {"user": "123"}, {"user": "darky"}, {"user": "darky 21"}, {"user": "darky darky"}, {"user": "darky abc efg"}],
            [None, None, {"user": 123}, None, None, None, None],
            [None, None, {"user": "123"}, {"user": "darky"}, {"user": "darky"}, {"user": "darky"}, {"user": "darky"}],
            [None, None, None, None, {"user": "darky", "age": 21}, None, None],
            [None, None, None, None, {"user": "darky", "testword": "21"}, {"user": "darky", "testword": "darky"}, {"user": "darky abc", "testword": "efg"}]
        ]
    }

@pytest.mark.asyncio
async def test_parse(twiml: TwiML, parse_dataset: dict):
    for _temp in range(6):
        twiml.update_template(parse_dataset.get("templates")[_temp])
        for _msg in range(7):
            result = await twiml.parse(parse_dataset.get("messages")[_msg])
            assert result == parse_dataset.get("results")[_temp][_msg]

@pytest.fixture
def mentions_dataset():
    return {
        "messages": [
            "",
            "test",
            "[id1|@durov]",
            "[id123|UserTest]",
            "test [id1|UserTest]",
            "[id1|UserTest] test",
            "[id1|UserTest1] [id2|UserTest2]",
            "[club1|@club1]",
            "[club123|GroupTest]",
            "test [club1|GroupTest]",
            "[club1|GroupTest] test",
            "[club1|GroupTest1] [club1|GroupTest2]",
            "[id1|User1] [club2|Group2] [id2|User2] [club1|Group1]"
        ],
        "results": [
            {"mentions": []},
            {"mentions": []},
            {"mentions": [{"type": "id", "id": 1, "screen_name": "id1", "text": "@durov"}]},
            {"mentions": [{"type": "id", "id": 123, "screen_name": "id123", "text": "UserTest"}]},
            {"mentions": [{"type": "id", "id": 1, "screen_name": "id1", "text": "UserTest"}]},
            {"mentions": [{"type": "id", "id": 1, "screen_name": "id1", "text": "UserTest"}]},
            {"mentions": [{"type": "id", "id": 1, "screen_name": "id1", "text": "UserTest1"}, {"type": "id", "id": 2, "screen_name": "id2", "text": "UserTest2"}]},
            {"mentions": [{"type": "club", "id": 1, "screen_name": "club1", "text": "@club1"}]},
            {"mentions": [{"type": "club", "id": 123, "screen_name": "club123", "text": "GroupTest"}]},
            {"mentions": [{"type": "club", "id": 1, "screen_name": "club1", "text": "GroupTest"}]},
            {"mentions": [{"type": "club", "id": 1, "screen_name": "club1", "text": "GroupTest"}]},
            {"mentions": [{"type": "club", "id": 1, "screen_name": "club1", "text": "GroupTest1"}, {"type": "club", "id": 1, "screen_name": "club1", "text": "GroupTest2"}]},
            {"mentions": [
                {"type": "id", "id": 1, "screen_name": "id1", "text": "User1"},
                {"type": "club", "id": 2, "screen_name": "club2", "text": "Group2"},
                {"type": "id", "id": 2, "screen_name": "id2", "text": "User2"},
                {"type": "club", "id": 1, "screen_name": "club1", "text": "Group1"}
            ]}
        ]
    }

@pytest.mark.asyncio
async def test_extract_mentions(twiml: TwiML, mentions_dataset: dict):
    for _msg in range(13):
        mentions = await twiml.extract_mentions(mentions_dataset.get("messages")[_msg])
        assert mentions == mentions_dataset.get("results")[_msg]


from typing import overload

from .base import BaseMethodsGroup

class Users(BaseMethodsGroup):

    @overload
    async def get(self,
                  user_ids: str,
                  fields: str = None,
                  name_case: str = None,
                  from_group_id: int = None):
        ...

    async def get(self,
                  user_ids: str | list[str],
                  fields: str = None,
                  name_case: str = None,
                  from_group_id: int = None):
        '''
        Метод позволяет получить информацию о пользователях.
        (см. https://dev.vk.ru/ru/method/users.get)
        '''
        values = {
            "user_ids": ",".join([f"{uid}" for uid in user_ids]) if isinstance(user_ids, list) else f"{user_ids}",
            "fields": fields,
            "name_case": name_case,
            "from_group_id": from_group_id,
            "v": self.__api_version__
        }
        response = await self.base_api.base_get_method(api_method=f"{self.method}.get",
                                                       values=values)
        return response


from .base import BaseMethodsGroup

class Groups(BaseMethodsGroup):
    
    async def getLongPollServer(self,
                                group_id:int=None) -> dict:
        
        '''
        Returns data for connection to Bots Longpoll API

        :param group_id: - Group's ID
        :type group_id: int | None
        '''

        values = {
            "group_id": abs(group_id) if group_id is not None else self.__group_id__,
            "v": self.__api_version__
        }
        response = await self.base_api.base_get_method(api_method=f"{self.method}.getLongPollServer",
                                                       values=values)
        return response
    


class TwiVKStates:

    INITIALIZING = "initializing"
    DISABLED = "disabled"
    STARTING = "starting"
    SHUTTING_DOWN = "shutting_down"
    READY = "ready"
    ERROR = "error"

class TwiAPIStates(TwiVKStates):
    pass


from typing import TYPE_CHECKING

from .framework.twilight_vk import TwilightVK

from .framework import (
    handlers,
    exceptions,
    rules
)

from .api.twilight_api import TwilightAPI

from .utils.config import CONFIG

from .logger.darky_logger import DarkyLogger
from .logger.darky_visual import (
    Visual,
    STYLE,
    BG,
    FG
)
from .logger.formatters import (
    DarkyConsoleFormatter,
    DarkyFileFormatter
)


import pytest
import asyncio

from twilight_vk import TwilightVK
from twilight_vk.framework.rules import TrueRule
from twilight_vk.framework.exceptions.framework import FrameworkError
from twilight_vk.utils.types.event_types import BotEventType

@pytest.fixture
def bot():
    botInst = TwilightVK(
        ACCESS_TOKEN="123",
        GROUP_ID=123
    )
    return botInst

class MockPolling:

    def get() -> dict:
        events = {"updates": []}
        for attr_name, attr_value in vars(BotEventType).items():
            if "__" not in attr_name:
                events["updates"].append({"type": attr_value, "object": {"message": {"peer_id": 123, "conversation_message_id": 1}}})
        return events

    
@pytest.mark.asyncio
async def test_labeler(bot: TwilightVK, caplog):

    @bot.on_event.all(TrueRule())
    async def handle(event: dict):
        return "OK"
    
    for record in caplog.records:
        if record.levelname == "INIT":
            assert record.message in ["Rule TrueRule() is initiated",
                                      "handle() was added to MESSAGE_NEW with rules: ['TrueRule']",
                                      "handle() was added to MESSAGE_REPLY with rules: ['TrueRule']",
                                      "handle() was added to DEFAULT_HANDLER with rules: ['TrueRule']"]
    for handler_name in bot.__bot__._router._handlers.keys():
        assert bot.__bot__._router._handlers[handler_name].__funcs__[0].get("func", False) == handle
        assert isinstance(bot.__bot__._router._handlers[handler_name].__funcs__[0].get("rules")[0], TrueRule)
    
@pytest.mark.asyncio
async def test_eventhandlers(monkeypatch):

    _bot = TwilightVK(ACCESS_TOKEN="123", GROUP_ID=123)
    
    @_bot.on_event.all(TrueRule())
    async def handle(event: dict):
        assert isinstance(event, dict)
        assert event.keys() == {'type': 'test', 'object': 'test'}.keys()
        return "OK"
    
    @_bot.on_event.all(TrueRule())
    async def handle2(event: dict):
        assert isinstance(event, dict)
        assert event.keys() == {'type': 'test', 'object': 'test'}.keys()
    
    async def fake_messageSend(*args, **kwargs):
        return True
    
    for handler_name in _bot.__bot__._router._handlers.keys():
        monkeypatch.setattr(_bot.__bot__._router._handlers[handler_name].vk_methods.messages, "send", fake_messageSend)

    results = await asyncio.gather(
        *(_bot.__bot__._router.route(event) for event in MockPolling.get()["updates"]),
        return_exceptions=False
    )
    for result in results:
        
        if result is None:
            assert True
            continue

        assert result


import asyncio
import contextlib
import logging
from typing import Coroutine, overload

class TwiTaskManager:

    def __init__(self, tasks: list[Coroutine] = [], loop: asyncio.AbstractEventLoop = None):
        '''
        Twilight's event loop wrapper.
        Allows to manage async tasks with creating a loop and running it forever

        :param tasks: List of coroutines should be ran on the start of the loop
        :type tasks: list[Coroutine]

        :param loop: Already created loop if it exist
        :type loop: asyncio.AbstractEventLoop
        '''
        self.logger = logging.getLogger(name="loop-manager")
        self._tasks: list[asyncio.Task] = tasks
        self._loop: asyncio.AbstractEventLoop = asyncio.get_event_loop() if loop is None else loop
    
    def get_loop(self):
        '''Returns a loop if it was created'''
        if self._loop:
            return self._loop
        self.logger.warning("Event loop was not created")
    
    def run(self, stop_when: str = asyncio.FIRST_EXCEPTION):
        '''
        Run the event loop until all task will be completed or first_exception will be caught

        :param stop_when: Let the loop manager know when it should stop (ALL_COMPLETED / FIRST_COMPLETED / FIRST_EXCEPTION)
        :type stop_when: str
        '''
        self.logger.debug(f"Loop was started with {len(self._tasks)} tasks")

        while self._tasks:
            self._loop.create_task(self._tasks.pop(0))
        
        _all_tasks = asyncio.all_tasks(self._loop)

        try:
            while _all_tasks:
                _done, _ = self._loop.run_until_complete(
                    asyncio.wait(_all_tasks, return_when=stop_when)
                )

                for _result in _done:
                    try:
                        _result.result()
                    except Exception as exc:
                        self.logger.critical(f"Task {_result.get_coro().__name__} retuned the exception: {exc}", exc_info=True)

                _all_tasks = asyncio.all_tasks(self._loop)

        except KeyboardInterrupt:
            self.logger.warning("KeyboardInterrupt recieved, shutting down...")
            self.stop(_all_tasks)

        finally:
            self.close()

    def stop(self, tasks: list[asyncio.Task] = None):
        '''
        Force stop with cancelling all the tasks

        :param tasks: List of currently created tasks in the event loop
        :type tasks: list[asyncio.Task]
        '''
        self.logger.debug("Stopping all tasks...")
        self.cancel_tasks(tasks)
        self.logger.debug("Tasks was stopped")

    def cancel_tasks(self, tasks: list[asyncio.Task] = None, targets: Coroutine | list[Coroutine] = None):

        if tasks is None:
            tasks = asyncio.all_tasks(self._loop)

        #TODO: Cancelling the certain tasks, not the all ones

        #if targets is None:
        #    tasks_to_cancel = asyncio.gather(*tasks)
        #else:
        #    _tasks_to_gather = set()
        #    for _task in tasks:
        #        for _target in targets:
        #            if _task.get_coro() == _target:
        #                _tasks_to_gather.add(_task)
        #    tasks_to_cancel = asyncio.gather(*_tasks_to_gather)
        
        tasks_to_cancel = asyncio.gather(*tasks)
        tasks_to_cancel.cancel()

        with contextlib.suppress(asyncio.CancelledError):
            self._loop.run_until_complete(
                asyncio.wait_for(
                    tasks_to_cancel,
                    timeout=10
                )
            )

    def close(self):
        '''
        Closing the current event loop
        '''
        if self._loop.is_running():
            self.logger.debug("Loop will be closed")
            self._loop.close()
    
    @overload
    def add_task(self, tasks: list[Coroutine]):
        ...
    
    @overload
    def add_task(self, task: Coroutine):
        ...
    
    def add_task(self, tasks: Coroutine | list[Coroutine]):
        '''
        Adds the task to the event loop

        :param tasks: Coroutine or list of coroutine functions to be added in event loop
        :type tasks: Coroutine | list[Coroutine]
        '''

        _tasks = tasks if isinstance(tasks, list) else [tasks]

        for _task in _tasks:
            if asyncio.iscoroutinefunction(_task):
                _task = _task()
            elif not asyncio.iscoroutine(_task):
                raise TypeError("Task should be coroutine or coroutine function")
            
            if self._loop and self._loop.is_running():
                self._loop.create_task(_task)
                self.logger.debug(f"Task {_task.__name__} was added to the running event loop")
            else:
                self._tasks.append(_task)
                self.logger.debug(f"Task {_task.__name__} was added to the event loop")



from aiohttp import ClientSession, ClientResponse, ClientTimeout, TCPConnector

class Http:

    def __init__(self,
                 headers:dict|None=None,
                 timeout:int=30):
        '''
        This class allows to use HTTP requests asynchronously
        '''
        self.session = None
        self.headers = headers
        self.timeout = timeout
    
    async def __getSession__(self):
        if self.session is None:
            self.session = ClientSession(headers=self.headers,
                                         timeout=ClientTimeout(self.timeout),
                                         connector=TCPConnector(force_close=True))
        
    @staticmethod
    async def __isRaw__(response:ClientResponse, raw:bool=False):
        if raw:
            return response
        return await response.json()

    async def get(self,
                  url:str,
                  params:dict|None=None,
                  headers:dict|None=None,
                  raw:bool=True) -> ClientResponse | dict:
        '''
        HTTP GET method

        :param url: The url to get the response from
        :type url: str

        :param params: Dictionary containing the optional data to be sent in the GET request
        :type params: dict | None

        :param headers: Optional dictionary of HTTP headers to include in the request
        :type headers: dict, optional

        :param raw: Defines the raw/json response
        :type raw: bool
        '''
        await self.__getSession__()
        response = await self.session.get(url=url,
                                          params=params,
                                          headers=headers)
        return await self.__isRaw__(response, raw=raw)
    
    async def post(self,
                   url:str,
                   data:dict,
                   params:dict={},
                   headers:dict=None,
                   raw:bool=True) -> ClientResponse | dict:
        '''
        HTTP POST method

        :param url: The URL to send the POST request to
        :type url: str

        :param data: Dictionary containing the data to be sent in the POST request body
        :type data: dict

        :param headers: Optional dictionary of HTTP headers to include in the request
        :type headers: dict, optional

        :param raw: Defines the raw/json response
        :type raw: bool
        '''
        await self.__getSession__()
        response = await self.session.post(url=url,
                                           params=params,
                                           json=data,
                                           headers=headers)
        return await self.__isRaw__(response, raw=raw)
    
    async def close(self):
        if self.session is not None and not self.session.closed:
            await self.session.close()
            self.session = None


from .framework import (
    FrameworkError,
    InitializationError
)

from .handler import (
    HandlerError,
    ResponseHandlerError
)

from .validator import (
    ValidationError,
    HttpValidationError,
    EventValidationError
)

from .vkapi import (
    LongPollError,
    VkApiError,
    AuthError
)


from typing import TYPE_CHECKING
from typing import Any
import logging

from ...utils.config import CONFIG

if TYPE_CHECKING:
    from ..methods import VkMethods

class BaseRule:

    def __init__(self, **kwargs):
        '''
        Base rule with base logic
        All child rules should inherit this one with changing check() function's logic

        :param kwargs: Dict of addictional arguments to this function
        :type kwargs: dict
        '''
        self.event: dict = None
        self.methods: "VkMethods" = None
        self.kwargs: dict = kwargs
        self.logger = logging.getLogger("rule-handler")
        self.__parseKwargs__()
        self.logger.log(1, f"Rule {self.__class__.__name__}({", ".join(self.__printKwargs__())}) is initiated")
    
    def __parseKwargs__(self):
        '''
        Parsing kwargs attribute, allowing to use each item as separate rule's attribute
        '''
        for key, value in self.kwargs.items():
            setattr(self, key, value)
    
    def __printKwargs__(self):
        output = []
        for key, value in self.kwargs.items():
            if key in ["rules", "rule"]:
                output.append(f"{key}: {", ".join(obj.__class__.__name__ for obj in value)}")
                continue
            output.append(f"{key}: {value}")
        return output

    def __getattr__(self, name):
        '''
        Allows to handle errors with parsing
        '''
        if name not in ['event', 'kwargs', 'methods', 'logger']:
            return getattr(self, self.kwargs[name])
    
    async def __linkVkMethods__(self, methods):
        '''
        Updates the methods attribute so you could make api requests from inside the rule
        '''
        try:
            self.logger.debug(f"Linking VkMethods class to the {self.__class__.__name__}...")
            self.methods = methods
        except Exception as ex:
            self.logger.error(f"Got an error while linking", exc_info=True)
            return False
    
    async def __check__(self, event: dict) -> bool:
        '''
        The shell for Rule.check()
        Allows to logging
        '''
        try:
            self.logger.debug(f"Checking rule {self.__class__.__name__}({self.__printKwargs__()})...")
            response = await self.check(event)
            self.logger.debug(f"Rule {self.__class__.__name__} returned {response}")
            return response
        except Exception as ex:
            self.logger.error(f"Rule {self.__class__.__name__} returned an exception", exc_info=True)
            return False

    async def check(self, event: dict) -> bool:
        '''
        Main function with specific check logic for specific rule.
        It may be different in different rules
        It should always return the boolean as the result
        '''
        pass

    def __and__(self, other: "BaseRule"):
        return AndRule(self, other)
    
    def and_(self, other: "BaseRule"):
        return AndRule(self, other)
    
    def __or__(self, other: "BaseRule"):
        return OrRule(self, other)
    
    def or_(self, other: "BaseRule"):
        return OrRule(self, other)
    
    def __not__(self):
        return NotRule(self)
    
    def not_(self):
        return NotRule(self)


class AndRule(BaseRule):
    
    def __init__(self, *rules: BaseRule):
        '''
        Модификатор AND для комбинирования правил
        '''
        super().__init__(
            rules = rules
        )

    async def check(self, event: dict):
        results = {}

        for rule in self.rules:

            result = await rule.__check__(event)

            if isinstance(result, Exception):
                return result
            
            if result is False:
                return False
            
            if isinstance(result, dict):
                results.update(result)

        if results != {}:
            return results
        
        return True

class OrRule(BaseRule):
    
    def __init__(self, *rules: BaseRule):
        '''
        Модификатор OR для комбинирования правил
        '''
        super().__init__(
            rules = rules
        )
    
    async def check(self, event: dict):

        for rule in self.rules:

            result = await rule.__check__(event)

            if isinstance(result, Exception):
                return result
            
            if result is True:
                return True
            
        return False

class NotRule(BaseRule):
    
    def __init__(self, rule: BaseRule):
        '''
        Модификатор NOT для комбинирования правил
        '''
        super().__init__(
            rule = rule
        )
    
    async def check(self, event: dict):

        result = await self.rule.__check__(event)
        return not result if isinstance(result, bool) else False


from aiohttp import ClientResponse
from .framework import FrameworkError

class ValidationError(FrameworkError):
    
    def __init__(self,
                 response: dict | ClientResponse ,
                 message: str | None = None):
        '''
        Базовый класс ошибки валидации

        :param response: Ответ HTTP запроса или тело ответа
        :type response: dict | ClientResponse
        '''
        self.response = response
        self.message = message
    
    def __str__(self):
        return f"Validation error"\
        f"{f" : {self.message}" if self.message is not None else ""}"

class HttpValidationError(ValidationError):
    
    def __init__(self,
                 isValid: bool,
                 isSuccess: bool,
                 response: ClientResponse):
        '''
        Ошибка валидации ответов от любых HTTP запросов
        Возникает при некорректном raw-формате ответа либо при кодах ответа которые не соответствуют успешным кодам

        :param isValid: Результат проверки RAW
        :type isValid: bool

        :param isSuccess: Результат проверки успешного кода в ответе
        :type isSuccess: bool

        :param response: Ответ от HTTP запроса
        :type response: ClientResponse
        '''
        self.isRaw = isValid
        self.isSuccess = isSuccess
        self.response = response
    
    def __str__(self):
        return f"Response validation error"\
        f"{": Is not valid raw " if self.isRaw is None else ""}"\
        f"{": Status code is not success " if self.isSuccess is None else ""}"

class EventValidationError(ValidationError):

    def __init__(self,
                 jsonIsValid: bool,
                 fieldsAreValid: bool,
                 content: dict):
        '''
        Ошибка валидации ответов от API
        Возникает при ошибках в ответе или при некорректном формате

        :param jsonIsValid: Результат проверки наличия JSON в теле ответа
        :type jsonIsValid: bool

        :param fieldsAreValid: Результат проверки обязательных полей в теле ответа
        :type fieldsAreValid: bool

        :param content: Тело ответа в формате JSON
        :type response: dict
        '''
        self.jsonIsValid = jsonIsValid
        self.fieldsAreValid = fieldsAreValid
        self.content = content

    def __str__(self):
        return f"Response validation error"\
        f"{": Content is not JSON " if self.jsonIsValid is None else ""}"\
        f"{": Response doesn't contain the required fields " if self.fieldsAreValid is None else ""}"


import dotenv, os

import twilight_vk
from twilight_vk.framework.rules import TextRule

dotenv.load_dotenv()

bot = twilight_vk.TwilightVK(
    BOT_NAME=os.getenv("BOT_NAME"),
    ACCESS_TOKEN=os.getenv("BOT_ACCESS_TOKEN"),
    GROUP_ID=os.getenv("BOT_GROUP_ID")
)

@bot.on_event.message_new(TextRule(value=["привет", "hello"], ignore_case=True))
async def hello(event: dict):
    return "Hello world"

bot.start()


import os
import logging

logger = logging.getLogger(__name__)
logger.setLevel("DEBUG")

class VisualMeta(type):

    '''Automatic attribute formatting'''

    def __getattribute__(cls, name):
        attr = super().__getattribute__(name)
        if name != 'prefix' and isinstance(attr, str) and attr.isdigit():
            return f"{cls.prefix}{attr}m"
        return attr

    def __getattr__(cls, name):
        logger.warning(f"Attribute '{name}' not found in '{cls.__name__}'")
        return f"<{cls.__name__}.{name}>"

class Visual(metaclass=VisualMeta):

    '''Styles for print functions and etc'''

    prefix = "\033["
    
    @staticmethod
    def ansi(silent:bool=False) -> None:
        if os.name == "nt":
            os.system('')
            if silent==False:
                logger.info(f"ANSI support initiated!")

    @staticmethod
    def hex_to_rgb(hex:str) -> tuple[int, int, int]:
        '''
        Converts HEX code to RGB

        :param hex: your hex code need to be converted to rgb code
        :type hex: str

        :returns: tuple of RGB integers (R, G, B)
        '''
        hex = hex.lstrip("#")

        if len(hex) not in (3, 6):
            raise ValueError(f"Invalid HEX code: #{hex}")
        
        if len(hex) == 3:
            hex = ''.join(c * 2 for c in hex)
            
        return tuple(int(hex[i:i+2], 16) for i in (0, 2, 4))
    
    @staticmethod
    def CUSTOM_COLOR(color: str = "#FFF") -> str:
        pass
    
    @staticmethod
    def GRADIENT(text: str | list[str], colors: list = ["#FFF"]) -> str:
        pass


class STYLE(Visual):
    RESET = "0"
    BOLD = "1"
    DIM = "2"
    ITALIC = "3"
    UNDERLINE = "4"
    BLINK = "5"
    NEGATIVE = "7"
    INVISIBLE = "8"
    CROSS = "9"

class FG(STYLE):
    BLACK = "30"
    RED = "31"
    GREEN = "32"
    YELLOW = "33"
    BLUE = "34"
    PURPLE = "35"
    CIAN = "36"
    WHITE = "37"

    @staticmethod
    def CUSTOM_COLOR(color:str="#FFFFFF") -> str:
        r, g, b = Visual.hex_to_rgb(color)
        return f"{Visual.prefix}{"3"}8;2;{r};{g};{b}m"
    
    @staticmethod
    def GRADIENT(text:str|list[str], colors:list=["#FFFFFF"]) -> str:

        '''
        Returns the prepared gradient text
        
        :param text: - input text to gradient
        :type text: str

        :param colors: - list of HEX colors for gradient
        :type colors: list
        '''

        if len(colors) == 0:
            return text

        if len(colors) == 1:
            return f"{FG.CUSTOM_COLOR(colors[0])}{text}{STYLE.RESET}"
        
        #handling multilined text(ansii art support)
        if "\n" in text or isinstance(text, list):
            if not isinstance(text, list):
                lines = text.split('\n')
            else:
                lines = text.copy()
            for line in range(len(lines)):
                lines[line] = FG.GRADIENT(lines[line], colors)
            return f"{'\n'.join(lines)}{STYLE.RESET}"

        rgb = [Visual.hex_to_rgb(color) for color in colors]

        segment = len(text) // (len(colors)-1) #HelloWorld (RED to BLUE) segment = 10 // (2-1) 10

        result = ""
        for i in range(len(colors) - 1): #grabs the pair of colors even if there as more of them

            start_r, start_g, start_b = rgb[i]
            end_r, end_g, end_b = rgb[i+1]

            for step in range(segment):
                if i * segment + step >= len(text):
                    break
                char = text[i * segment + step]
                r = int(start_r + (end_r - start_r) * (step / segment))
                g = int(start_g + (end_g - start_g) * (step / segment))
                b = int(start_b + (end_b - start_b) * (step / segment))
                color_code = f"{Visual.prefix}{"3"}8;2;{r};{g};{b}m"
                result += f"{color_code}{char}"

        return f"{result}{STYLE.RESET}"

class BG(STYLE):
    BLACK = "40"
    RED = "41"
    GREEN = "42"
    YELLOW = "43"
    BLUE = "44"
    PURPLE = "45"
    CIAN = "46"
    WHITE = "47"

    @staticmethod
    def CUSTOM_COLOR(color:str="#FFFFFF") -> str:
        r, g, b = Visual.hex_to_rgb(color)
        return f"{Visual.prefix}{"4"}8;2;{r};{g};{b}m"
    
    @staticmethod
    def GRADIENT(text:str|list[str], colors:list=["#FFFFFF"]) -> str:

        '''
        Returns the prepared gradient text
        
        :param text: - input text to gradient
        :type text: str

        :param colors: - list of HEX colors for gradient
        :type colors: list
        '''

        if len(colors) == 0:
            return text

        if len(colors) == 1:
            return f"{BG.CUSTOM_COLOR(colors[0])}{text}{STYLE.RESET}"
        
        #handling multilined text(ansii art support)
        if "\n" in text or isinstance(text, list):
            if not isinstance(text, list):
                lines = text.split('\n')
            else:
                lines = text.copy()
            for line in range(len(lines)):
                lines[line] = BG.GRADIENT(lines[line], colors)
            return f"{'\n'.join(lines)}{STYLE.RESET}"

        rgb = [Visual.hex_to_rgb(color) for color in colors]

        segment = len(text) // (len(colors)-1) #HelloWorld (RED to BLUE) segment = 10 // (2-1) 10

        result = ""
        for i in range(len(colors) - 1): #grabs the pair of colors even if there as more of them

            start_r, start_g, start_b = rgb[i]
            end_r, end_g, end_b = rgb[i+1]

            for step in range(segment):
                if i * segment + step >= len(text):
                    break
                char = text[i * segment + step]
                r = int(start_r + (end_r - start_r) * (step / segment))
                g = int(start_g + (end_g - start_g) * (step / segment))
                b = int(start_b + (end_b - start_b) * (step / segment))
                color_code = f"{Visual.prefix}{"4"}8;2;{r};{g};{b}m"
                result += f"{color_code}{char}"

        return f"{result}{STYLE.RESET}"


from typing import TYPE_CHECKING
import logging
import asyncio

from .event_handlers import (
    BASE_EVENT_HANDLER,
    DEFAULT_HANDLER,
    MESSAGE_NEW,
    MESSAGE_REPLY
)
from ...utils.types.event_types import BotEventType

if TYPE_CHECKING:
    from ..methods import VkMethods
    from ...utils.event_loop import TwiTaskManager

class OnEventLabeler:

    def __init__(self, handlers: dict):
        self.handlers = handlers

    def all(self, *rules):
        def decorator(func):
            for handler_name in self.handlers.keys():
                self.handlers[handler_name].__add__(func, rules)
            return func
        return decorator
    
    def message_new(self, *rules):
        def decorator(func):
            self.handlers[BotEventType.MESSAGE_NEW].__add__(func, rules)
            return func
        return decorator

class EventRouter:

    def __init__(self,
                 vk_methods:'VkMethods',
                 _loop_wrapper: 'TwiTaskManager'):
        '''
        Router for events.
        Allows to route events to separate event handlers
        '''
        self.vk_methods = vk_methods
        self._loop_wrapper = _loop_wrapper
        self.logger = logging.getLogger("event-router")
        self._handlers = {
            "default": DEFAULT_HANDLER(self.vk_methods),
            BotEventType.MESSAGE_NEW: MESSAGE_NEW(self.vk_methods),
            BotEventType.MESSAGE_REPLY: MESSAGE_REPLY(self.vk_methods)
        }
        self.on_event = OnEventLabeler(self._handlers)
    
    async def handle(self, polling_response: dict):
        '''
        Handles the event list

        :param polling_response: Response from the polling requests contains "ts" and "updates" keys
        :type polling_response: dict
        '''
        self.logger.debug("Routing the events...")
        events = []
        for event in polling_response["updates"]:
            events.append(
                self._loop_wrapper._loop.create_task(
                    self.route(event)
                )
            )
        try:
            await asyncio.gather(*events, return_exceptions=False)
        except Exception as exc:
            self.logger.error(f"{exc.__class__.__name__}: {exc}", exc_info=True)

    async def route(self, current_event: dict):
        '''
        Routing the event to the exact handler
        '''
        event_type = current_event.get("type", "default")
        handler:BASE_EVENT_HANDLER = self._handlers.get(event_type, self._handlers["default"])
        self.logger.debug(f"Routing the event {current_event.get("type")} to the {handler.__class__.__name__} handler")
        await handler.__executeAsync__(current_event)


import pytest

from twilight_vk import TwilightVK

class MockFakeApiResponse:
    
    def get():
        return {
            "response": {
                "status": "ok"
            }
        }
    
@pytest.mark.asyncio
async def test_base_api_get(monkeypatch):
    bot = TwilightVK(
        ACCESS_TOKEN="123",
        GROUP_ID=123
    )

    async def fake_base_get_method(
            api_method: str,
            values: dict = {},
            headers: dict = {}
    ):
        return MockFakeApiResponse.get()
    
    monkeypatch.setattr(bot.__bot__.base_methods, "base_get_method", fake_base_get_method)
    
    assert await bot.__bot__.base_methods.base_get_method("testmethod/",
                                                          {"key": "value"},
                                                          headers={"Content-Type": "application/json"}) == MockFakeApiResponse.get()


import re
import logging
from typing import Literal
from copy import copy

from uvicorn.logging import AccessFormatter

from .darky_visual import STYLE, FG, BG


class MaskData:

    def mask_sensitive(message, keys=None):
        if keys is None:
            keys = ["access_token", "key", "Authorization"]
        key_pattern = "|".join(keys)
        
        patterns = [
            rf'([\'"])({key_pattern})\1\s*:\s*({"|".join([r'([\'"]).*?\1', r'([0-9])*'])})',
            rf'({key_pattern})=.*?(&|$)'
        ]
        combined_pattern = '|'.join(patterns)

        def replacer(match):
            if match.group(1):
                key_quote = match.group(1)
                key = match.group(2)
                if match.group(4):
                    value_quote = match.group(4)
                    return f'{key_quote}{key}{key_quote}: {value_quote}***{value_quote}'
                else:
                    return f"{key_quote}{key}{key_quote}: ***"
            else:
                key = match.group(6)
                suffix = match.group(7)
                return f'{key}=***{suffix}'
        
        try:
            return re.sub(combined_pattern, replacer, message, flags=re.IGNORECASE)
        except re.error as e:
            return f"Regex error: {e} --- Message: {message}"


class DarkyConsoleFormatter(logging.Formatter):

    levelname_colors = {
        "SUBDEBUG": f"{FG.CUSTOM_COLOR("#448")}SUBDEBUG{STYLE.RESET}",
        "DEBUG": f"{FG.CUSTOM_COLOR("#66A")}DEBUG{STYLE.RESET}",
        "INFO": f"{FG.CUSTOM_COLOR("#16F")}INFO{STYLE.RESET}",
        "WARNING": f"{FG.YELLOW}WARNING{STYLE.RESET}",
        "ERROR": f"{FG.BOLD}{FG.RED}ERROR{STYLE.RESET}",
        "CRITICAL": f"{FG.BOLD}{BG.RED}{FG.WHITE}CRITICAL{STYLE.RESET}",
        "NOTE": f"{FG.CUSTOM_COLOR("##AAA")}NOTE{STYLE.RESET}"
    }

    default_levelname_color = f"{FG.CUSTOM_COLOR("#555")}%s{STYLE.RESET}"
    twiname_color = f"{FG.CUSTOM_COLOR("#22B")}%s{STYLE.RESET}"
    asctime_color = f"{FG.CUSTOM_COLOR("#DDD")}%s{STYLE.RESET}"
    name_color = f"{FG.CUSTOM_COLOR("#555")}%s{STYLE.RESET}"

    def __init__(
            self,
            fmt: str | None = None,
            datefmt: str | None = None,
            style: Literal["%", "{", "$"] = "%",
            colored: Literal[True, False] = False,
            color_core_name: Literal[True, False] = False
    ):
        '''
        Initializes the console formatter for logging module

        :param fmt: Log message format
        :type fmt: str | None

        :param datefmt: Time format
        :type datefmt: str | None

        :param style: Format style
        
        :param colored: Sets the colors for log message
        :type colored: bool
        '''
        self.colored = colored
        self.color_core_name = color_core_name
        super().__init__(fmt=fmt, datefmt=datefmt, style=style)
    
    def color_levename(self, levelname: str) -> str:
        '''
        Applies colors to the levelname
        '''
        if self.colored:
            return self.levelname_colors.get(levelname, self.default_levelname_color % levelname)
        return levelname
    
    def formatTime(self, record, datefmt = None):
        '''
        Applies color to the asctime
        '''
        time_str = super().formatTime(record, datefmt)
        if self.colored:
            return self.asctime_color % time_str
        return time_str

    def formatException(self, ei):
        '''
        Applies colors to the exception's traceback
        '''
        if self.colored:
            return f"{FG.RED}{super().formatException(ei)}{STYLE.RESET}"
        return super().formatException(ei)

    def format(self, record):
        '''
        Formatting the log message
        '''
        record_copy = copy(record)
        record_copy.msg = MaskData.mask_sensitive(record_copy.msg)
        if self.colored:
            if self.color_core_name and "twilight" in record_copy.name:
                #record_copy.name = f"{STYLE.GRADIENT(f"{record_copy.name}", ["#44F", "#A6F"])}{STYLE.RESET}"
                record_copy.name = self.twiname_color % record_copy.name
            else:
                record_copy.name = self.name_color % record_copy.name
            if record_copy.levelname == "CRITICAL":
                record_copy.msg = f"{FG.RED}{record_copy.msg}{STYLE.RESET}"
            record_copy.levelname = self.color_levename(record_copy.levelname)
        record_copy.name = record_copy.name + " " * (15 - len(record.name))
        record_copy.levelname = record_copy.levelname + " " * (8 - len(record.levelname))
        return super().format(record_copy)
        

class DarkyFileFormatter(logging.Formatter):

    def __init__(
            self,
            fmt: str | None = None,
            datefmt: str | None = None,
            style: Literal["%", "{", "$"] = "%",
    ):
        '''
        Initializes the console formatter for logging module

        :param fmt: Log message format
        :type fmt: str | None

        :param datefmt: Time format
        :type datefmt: str | None

        :param style: Format style
        '''
        super().__init__(fmt=fmt, datefmt=datefmt, style=style)
    
    def formatTime(self, record, datefmt = None):
        return super().formatTime(record, datefmt)
    
    def formatException(self, ei):
        return super().formatException(ei)

    def format(self, record):
        '''
        Removes the colors for file logging
        '''
        record_copy = copy(record)
        record_copy.msg = MaskData.mask_sensitive(record_copy.msg)
        record_copy.levelname =      re.sub(r'\033\[.*?m', '', record_copy.levelname)
        record_copy.msg =            re.sub(r'\033\[.*?m', '', record_copy.msg)
        record_copy.name =           re.sub(r'\033\[.*?m', '', record_copy.name)

        record_copy.name = record_copy.name + " " * (15 - len(record.name))
        record_copy.levelname = record_copy.levelname + " " * (8 - len(record.levelname))

        return super().format(record_copy)

class UvicornAccessFormatter(DarkyConsoleFormatter, AccessFormatter):
    pass


import re
from typing import TYPE_CHECKING
import logging
import asyncio

from fastapi import APIRouter

from ....utils.config import CONFIG
from ....utils.types.twi_states import TwiVKStates
from .vk_api import VkApiRouter

if TYPE_CHECKING:
    from ...twilight_vk import TwilightVK

class FrameworkRouter:

    def __init__(self, bot: 'TwilightVK'):
        '''
        Framework API router

        :param bot: TwilightVK object
        :type bot: TwilightVK
        '''
        self.logger = logging.getLogger("twi-api-fw")
        self.logger.log(1, f"Framework API Router was initiated")
        self.bot = bot

        self.router = APIRouter(
            tags=[bot.bot_name],
            prefix=f"/{re.escape(bot.bot_name)}"
        )

        self.router.add_api_route("/ping", self.ping, methods=["GET"],
                                  name="Ping Framework API",
                                  description="Pings the frameworks's main API router")
        self.router.add_api_route("/stop", self.stop, methods=["GET"],
                                  name="Stop Current Bot",
                                  description="Stops current linked bot")
        
        self.logger.debug(f"Including VkMethods router...")
        self.methods_router = VkApiRouter(self.bot)
        self.router.include_router(self.methods_router.get_router())
        self.logger.debug(f"Done")
    
    async def ping(self) -> dict:
        '''
        API METHOD
        Pings the bot
        '''
        if self.bot._state == TwiVKStates.READY:
            return {"response": {"message": "Pong OwO"}}
        
        return {"error": {"bot_state": f"{self.bot._state}"}}
    
    async def stop(self, force: bool = False) -> dict:
        '''
        API METHOD
        Stops the bot

        :param force: Forced stop
        :type force: bool
        '''
        if self.bot._state in [TwiVKStates.READY, TwiVKStates.STARTING]:
            self.bot.stop(force)

            while not self.bot._state == TwiVKStates.DISABLED:
                await asyncio.sleep(1.0)

            return {"response": {"message": "Bot was stopped", "forced": force}}
        
        return {"error": {"message": "Unable to stop the bot", "bot_state": f"{self.bot._state}"}}

    
    def get_router(self) -> APIRouter:
        return self.router


from .rules import (
    FalseRule,
    TrueRule,
    ContainsRule,
    TextRule,
    TwiMLRule,
    MentionRule,
    IsMentionedRule,
    ReplyRule,
    ForwardRule,
    AdminRule,
    IsAdminRule,
    InvitedRule,
    IsInvitedRule
)
from .base import BaseRule


import pytest
import asyncio
import pytest_asyncio

from twilight_vk.utils.event_loop import TwiTaskManager

@pytest_asyncio.fixture
def loop(_function_event_loop):
    wrapper = TwiTaskManager(loop = _function_event_loop)
    yield wrapper

@pytest.fixture
def log_fixture(caplog):
    caplog.set_level(0)
    return caplog

async def fake_coro1(wrapper: TwiTaskManager):
    wrapper.logger.info("Hello")

async def fake_coro2(wrapper: TwiTaskManager):
    wrapper.logger.info("World")

async def fake_coro3():
    await asyncio.sleep(5)

async def fake_coro4(wrapper: TwiTaskManager):
    wrapper.logger.info("Test1")

async def fake_coro5(wrapper: TwiTaskManager):
    wrapper.logger.info("Test2")\
    
async def fake_coro6():
    await asyncio.sleep(5)

def test_wrapper_run(loop: TwiTaskManager, log_fixture):
    loop.add_task(fake_coro1(loop))
    loop.run()
    assert "Hello" in log_fixture.text

@pytest.mark.asyncio
async def test_wrapper_addtask(loop: TwiTaskManager, log_fixture):
    loop.add_task(fake_coro2(loop))
    await asyncio.sleep(0.1)
    assert "World" in log_fixture.text

@pytest.mark.asyncio
async def test_wrapper_addtask_list(loop: TwiTaskManager, log_fixture):
    loop.add_task([fake_coro4(loop), fake_coro5(loop)])
    await asyncio.sleep(0.1)
    assert "Test1" in log_fixture.text
    assert "Test2" in log_fixture.text

def test_wrapper_stop(loop: TwiTaskManager, log_fixture):
    loop.add_task(fake_coro3())
    loop.stop()
    assert "Stopping all tasks..." in log_fixture.text
    assert "Tasks was stopped" in log_fixture.text


from typing import TYPE_CHECKING, AnyStr

from .framework import FrameworkError

if TYPE_CHECKING:
    from ...utils.types.response import ResponseHandler

class HandlerError(FrameworkError):

    def __init__(self,
                 message:str):
        '''
        Базовое исключение обработчика

        :param message: Опционально. Дополнительная информация об ошибке
        :type message: str
        '''
        self.message = message
    
    def __str__(self):
        return f"Handler error"\
        f"{f" : {self.message}" if self.message is not None else ""}"
    

class ResponseHandlerError(HandlerError):

    def __init__(self,
                 callback: "ResponseHandler" | AnyStr,
                 _isinstance: bool = None):
        '''
        Исключение обработчика ответов от функций

        :param callback: Ответ от функции, которая выполнилась в обработчике событий
        :type callback: ResponseHandler | Any
        
        :param instance_needed: Класс который должен был передаться в функции (всегда должен быть ResponseHandler)
        :type instance_needed: ResponseHandler
        '''
        self.callback = callback
        self._isinstance = _isinstance
    
    def __str__(self):
        return f"Response handler error"\
        f"{f" : function callback is not instance of ResponseHandler, make sure you are returning correct values in your functions"\
           if self._isinstance is not None and not self._isinstance else self.callback}"


import logging

from .groups import Groups
from .messages import Messages
from .users import Users
from ...utils.config import CONFIG

class VkMethods:

    def __init__(self,
                 baseMethods:object):
        self.logger = logging.getLogger("vk-methods")
        self.groups = Groups(baseMethods)
        self.messages = Messages(baseMethods)
        self.users = Users(baseMethods)


import logging

from aiohttp import ClientResponse
from http import HTTPStatus

from ..exceptions.validator import HttpValidationError

class HttpValidator:

    def __init__(self):
        self.logger = logging.getLogger("http-validator")

    async def __isValid__(self,
                          response:ClientResponse) -> bool:
        self.logger.debug(f"Type matching check...")
        if type(response) != ClientResponse:
            self.logger.error(f"Types does not match: {type(response)} != {ClientResponse}")
            return False
        return True
    
    async def __isSuccess__(self,
                            response:ClientResponse) -> bool:
        self.logger.debug(f"HTTP Status check...")
        if response.status != HTTPStatus.OK:
            self.logger.error(f"HTTP status is {response.status}")
            return False
        return True


    async def validate(self,
                       response:ClientResponse) -> ClientResponse:
        self.logger.debug(f"Validating response for [{response.method}] \"{response.url}\"...")

        isValid = await self.__isValid__(response)
        isSuccess = await self.__isSuccess__(response)

        if isValid and isSuccess:
            self.logger.debug(f"HTTP Response is valid")
            return response
        
        self.logger.warning(f"HTTP Response is not valid")
        self.logger.warning(f'{"isValid":<10}|{"":>1}{"isSuccess":<10}|{"":>1}{"response":<10}')
        self.logger.warning(f'{isValid:<10}|{"":>1}{isSuccess:<10}|{"":>1}{response}')
        
        raise HttpValidationError(isValid, isSuccess, response)


from .framework import FrameworkError

class LongPollError(FrameworkError):

    def __init__(self,
                 failed_code:int,
                 new_ts:int=None):
        '''
        Исключение при ошибках в ответ на запросы к LongPoll серверу

        :param failed_code: Код ошибки который возвращает сервер
        :type failed_code: int

        :param new_ts: Новый ts возвращаемый при failed: 1
        :type new_ts: int
        '''
        self.failed_code = failed_code
        self.new_ts = new_ts
        self.failed_msgs = {
            1: "The event history is outdated or has been partially lost. Use actual \"ts\"",
            2: "The key expired. New \"key\" is needed.",
            3: "The information is lost. New \"key\" and \"ts\" is needed."
        }

    def __str__(self):
        return f"[{self.failed_code}] {self.failed_msgs.get(self.failed_code)}"



class VkApiError(FrameworkError):
    
    def __init__(self,
                 error_code:int,
                 error_msg:str,
                 request_params:list
                 ):
        '''
        Исключение при ошибках в запросах к VK API

        :param error_code: Код ошибки
        :type error_code: int

        :param error_msg: Сообщение ошибки
        :type error_msg: str

        :param request_params: Переданные с запросом HTTP параметры
        :type request_params: list
        '''
        self.error_code = error_code
        self.error_msg = error_msg
        self.request_params = request_params
    
    def __str__(self):
        return f"[{self.error_code}] {self.error_msg}"

class AuthError(VkApiError):
    '''Ошибка авторизации в API'''
    pass


import logging

from aiohttp.client_exceptions import ClientConnectorDNSError
from aiohttp import ClientResponse

from .base_longpoll import BaseLongPoll
from ..exceptions.vkapi import (
    AuthError,
    VkApiError,
    LongPollError
)
from ...utils.config import CONFIG
from ..methods.base import VkBaseMethods
from ..methods import VkMethods
from ..handlers import EventRouter
from ..validators.http_validator import HttpValidator
from ..validators.event_validator import EventValidator
from ...http.async_http import Http
from ...utils.event_loop import TwiTaskManager

class BotsLongPoll(BaseLongPoll):

    def __init__(self,
                 access_token:str=None,
                 group_id:int=None,
                 api_version:str=CONFIG.VK_API.version,
                 loop_wrapper: TwiTaskManager = None) -> None:
        '''
        BotLongPoll provides a communication between your app and VK API
        Bots Long Poll API allows you to process community events in real time

        :param access_token: your group's access token, you can find it here(https://dev.vk.com/ru/api/access-token/community-token/in-community-settings)
        :type access_token: str

        :param group_id: your group's id, you can find it in your group's settings
        :type group_id: int

        :param api_version: version of VK API, by default its grabbed from DarkyVK's configuration
        :type api_version: str | None
        '''
        self.__stop__ = False
        self.authorized = False
        self.httpValidator = HttpValidator()
        self.eventValidator = EventValidator()
        self.httpClient = Http()
        self.logger = logging.getLogger(self.__class__.__name__.lower())
        self.__access_token__ = access_token
        self.__group_id__ = group_id
        self.__api_url__ = CONFIG.VK_API.url
        self.__api_version__ = api_version
        self.__server__ = None
        self.__key__ = None
        self.__ts__ = None
        self.__wait__ = CONFIG.VK_API.wait
        self.base_methods = VkBaseMethods(self.__api_url__, self.__access_token__, self.__group_id__)
        self.vk_methods = VkMethods(self.base_methods)
        self._loop_wrapper = loop_wrapper if loop_wrapper is not None else TwiTaskManager()
        self._router = EventRouter(self.vk_methods, _loop_wrapper=self._loop_wrapper)
        self.on_event = self._router.on_event
    
    async def auth(self, update_ts: bool = True):

        '''
        Authorizing framework in VK API by getting LongPoll server

        :param update_ts: Update the ts value flag
        :type update_ts: bool
        '''

        try:
            self.logger.debug(f"Authorization...")

            self.logger.debug(f"Getting Bots LongPoll Server...")
            response = await self.vk_methods.groups.getLongPollServer()
            response = response["response"]
            self.logger.debug(f"Server aquired: {response}")

            self.__key__ = response["key"]
            self.__server__ = response["server"]
            if update_ts:
                self.__ts__ = response["ts"]
            self.authorized = True

            self.logger.debug(f"Authorized")

        except ClientConnectorDNSError as ex:
            self.logger.critical(f"{ex}", exc_info=True)
            self.stop(_from="Authorization")
        except AuthError as ex:
            self.logger.error(f"Authrization error: {ex}")
            self.stop(_from="Authorization")
        except VkApiError as ex:
            self.logger.critical("Request to VK API was handled with error", exc_info=True)
            self.stop(_from="Authorization")

    
    async def check_event(self) -> dict:
        
        '''
        Getting the event from BotsLongPoll server
        '''
        try:
            self.logger.debug(f"Listening the BotsLongPoll server for events...")
            response = await self.httpClient.get(url=f"{self.__server__}",
                                            params={
                                                "act": "a_check",
                                                "key": self.__key__,
                                                "ts": self.__ts__,
                                                "wait": self.__wait__
                                            },
                                            raw=True)
            self.logger.debug(f"Validating...")
            response = await self.httpValidator.validate(response)
            response = await self.eventValidator.validate(response, from_polling=True)
            self.logger.debug(f"Got an event: {response}")

            if "ts" not in response:
                self.auth()

            self.__ts__ = response["ts"]
            
            return response
        except LongPollError as ex:
            match ex.failed_code:
                case 1:
                    self.logger.warning(f"The event history is outdated or has been partially lost. "\
                                        f"The application can receive events further using the new \"ts\" value from the response.")
                    self.__ts__ = ex.new_ts
                case 2:
                    self.logger.warning(f"The key is expired. Getting new one...")
                    await self.auth(update_ts=False)
                case 3:
                    self.logger.error(f"The information is lost. Reauthorizing...")
                    await self.auth()
            self.logger.debug(f"Retrying to get event...")
            if not self.__stop__:
                return await self.check_event()
        except VkApiError as ex:
            self.logger.critical("Request to VK API was handled with error", exc_info=True)
            self.stop()


import json

class ResponseHandler:

    def __init__(self,
                 peer_ids: int | list[int] = None,
                 domain: str = None,
                 chat_id: int = None,
                 message: str = None,
                 lat: str = None,
                 long: str = None,
                 attachment: str | list[str] = None,
                 reply_to: int = None,
                 forward_messages: int | list[int] = None,
                 forward: dict = None,
                 sticker_id: int = None,
                 keyboard: object = None,
                 template: object = None,
                 payload: object = None,
                 content_source: dict = None,
                 dont_parse_links: bool = None,
                 disable_mentions: bool = None):
        '''
        Allows to send responses for the bot


        '''
        self.peer_ids = ",".join([f"{peer_id}" for peer_id in peer_ids]) if isinstance(peer_ids, list) else f"{peer_ids}"
        self.domain = domain
        self.chat_id = chat_id
        self.message = message

        self.lat = lat
        self.long = long

        self.attachment = attachment = ",".join([f"{attach}" for attach in attachment]) if attachment is not None else None

        self.reply_to = reply_to
        self.forward_messages = forward_messages
        self.forward = json.dumps(forward) if forward is not None else None

        self.sticker_id = sticker_id
        
        self.keyboard = keyboard
        self.template = template
        self.payload = payload
        self.content_source = json.dumps(content_source) if content_source is not None else None
        self.dont_parse_links = dont_parse_links
        self.disable_mentions = disable_mentions
    
    def getData(self) -> dict:
        '''
        Returns all data from initialized class
        '''
        return self.__dict__


from typing import Callable, Optional
import logging.config, logging

from .darky_visual import Visual

class DarkyLogger:
    config = {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "console": {
                "format": "%(levelname)s - %(message)s"
            }
        },
        "handlers": {
            "console": {
                "level": "INITDEBUG",
                "class": "logging.StreamHandler",
                "formatter": "console"
            }
        },
        "loggers": {
            "test": {
                "handlers": ["console"],
                "level": "INITDEBUG",
                "propagate": False
            }
        }
    }

    def __init__(self, logger_name:str=None, configuration:dict=None, ansi:bool=True, silent:bool=False) -> None:

        r'''
        Класс DarkyLogger позволяет удобно и быстро инициализировать работу логгера logging

        :param logger_name: используется для присвоения уникального имени логгеру\
        (видно при использовании в форматировании %(name)s)
        :type logger_name: str 
        
        :param configuration: позволяет гибко настроить конфигурацию логгера
        (см. https://docs.python.org/3/library/logging.config.html#configuration-dictionary-schema)
        :type configuration: dict
        '''
        
        if logger_name is None:
            logger_name = __name__

        if configuration is None:
            configuration = DarkyLogger.config
        
        if ansi:
            Visual.ansi()

        logging.INITDEBUG = 1
        logging.addLevelName(logging.INITDEBUG, "INIT")

        logging.SUBDEBUG = 5
        logging.addLevelName(logging.SUBDEBUG, "SUBDEBUG")

        logging.NOTE = 100
        logging.addLevelName(logging.NOTE, "NOTE")

        logging.config.dictConfig(configuration)
        self.__logger__ = logging.getLogger(logger_name)
        
        self.initdebug: Callable[[str]] = \
            lambda msg, *args, **kwargs: self.__logger__.log(level=logging.INITDEBUG, msg=msg, *args, **kwargs)
        
        self.note: Callable[[str]] = \
            lambda msg, *args, **kwargs: self.__logger__.log(level=logging.NOTE, msg=msg, *args, **kwargs)
        
        self.subdebug: Callable[[str]] = \
            lambda msg, *args, **kwargs: self.__logger__.log(level=logging.SUBDEBUG, msg=msg, *args, **kwargs)
        
        self.debug: Callable[[str]] = \
            lambda msg, *args, **kwargs: self.__logger__.debug(msg, *args, **kwargs)
        
        self.info: Callable[[str]] = \
            lambda msg, *args, **kwargs: self.__logger__.info(msg, *args, **kwargs)
        
        self.warning: Callable[[str, bool]] = \
            lambda msg, *args, exc_info=False, **kwargs: self.__logger__.warning(msg, *args, exc_info=exc_info, **kwargs)
        
        self.error: Callable[[str, bool]] = \
            lambda msg, *args, exc_info=False, **kwargs: self.__logger__.error(msg, *args, exc_info=exc_info, **kwargs)
        
        self.critical: Callable[[str, bool]] = \
            lambda msg, *args, exc_info=False, **kwargs: self.__logger__.critical(msg, *args, exc_info=exc_info, **kwargs)
        
        if not silent:
            self.initdebug(f"DarkyLogger initiated")
    
    def __getattr__(self, name):

        if name not in ["initdebug", "subdebug", "debug", "info", "warning", "error", "critical", "note"]:
            self.__logger__.warning(f"'DarkyLogger' has no attribute '{name}'. \"INFO\" is used instead.")
            name = "info"
        attr = getattr(self.__logger__, name)
        return attr
    
    def get_logger(self) -> logging.Logger:

        '''
        Клонирует уже созданный ранее логгер для его использования в других модулях
        '''

        self.__logger__.debug(f"Initiated DarkyLogger has been requested")
        return self.__logger__


import re

class TwiML:

    def __init__(self, template: str=None):
        '''
        Twilight Markup Language for extracting args and mentions from message text

        :param template: Regex template for parsing
        :type template: str
        '''
        self.template = template
        self.placeholder_types = {
            "any": (r".+", str),
            "int": (r"\d+", int),
            "word": (r"\S+", str)
        }

    def update_template(self, template: str):
        '''
        Updates the template for parsing
        '''
        self.template = template

    async def parse(self,
                    message: str) -> dict:
        '''
        Parsing message for variable values.
        
        :param message: Message text
        :type message: str
        '''
        args = {}
        pattern = re.escape(self.template)
        placeholders = re.findall(r"<(\w+)(?::(\w+))?>", self.template)

        for name, type in placeholders:
            _type = type or "any"
            _regex, _convert_to = self.placeholder_types.get(_type)
            pattern = pattern.replace(
                re.escape(f"<{name}{":" + _type if type else ""}>"), f"({_regex})"
            )
            args[name] = _convert_to
        
        match = re.match(pattern, message)
        if not match:
            return None
        
        result = {}
        for (name, _), value in zip(placeholders, match.groups()):
            result[name] = args[name](value)
        return result
    
    async def extract_mentions(self,
                               message: str) -> dict:
        '''
        Extracts the mentions from message

        :param message: Message text
        :type message: str
        '''
        pattern = r"\[([a-zA-Z]+)(\d+)\|([^\]]*)\]"
        mentions = re.findall(pattern, message)

        result = {
            "mentions": [{
                "type": mention[0],
                "id": int(mention[1]),
                "screen_name": f"{mention[0]}{mention[1]}",
                "text": mention[2]
            } for mention in mentions]
        }

        return result


from twilight_vk.logger.darky_logger import DarkyLogger
from twilight_vk.logger.formatters import (
    DarkyConsoleFormatter,
    DarkyFileFormatter
)

CONFIG = {
    "version": 1,
    "disable_existing_loggers": True,
    "formatters": {
        "file": {
            "()": DarkyFileFormatter,
            "fmt": "%(name)s | %(asctime)s | %(levelname)s | %(message)s"
        },
        "console": {
            "()": DarkyConsoleFormatter,
            "fmt": "%(name)s | %(asctime)s | %(levelname)s | %(message)s",
            "colored": True
        }
    },
    "handlers": {
        "file": {
            "class": "logging.handlers.RotatingFileHandler",
            "level": "INIT",
            "formatter": "file",
            "filename": "tests/test_logging.log",
            "encoding": "utf8",
            "backupCount": 1
        },
        "console": {
            "class": "logging.StreamHandler",
            "level": "INIT",
            "formatter": "console"
        }
    },
    "loggers": {
        "test": {
            "handlers": ["console", "file"],
            "level": "INIT",
            "propagate": True
        }
    }
}

def test_logging(caplog):

    print()
    logger = DarkyLogger("test", CONFIG, silent=True)

    for level in [logger.initdebug, logger.note, logger.subdebug, logger.debug, logger.info, logger.warning, logger.error]:
        level(f"Mlem")
        level("{'key': 123}, {'access_token': 'abc'}, {\"access_token\": \"abc\"}")
        level(f"https://mlem.api/mlem?access_token=123")
        level(f"https://mlem.api/mlem?access_token=123&abs=True")
    logger.mlem("Mlem")
    try:
        raise Exception
    except Exception as ex:
        logger.critical(f"We got an error! Mlem", exc_info=True)
    
    for record in caplog.records:
        if record.message not in ["ANSI support initiated!", "DarkyLogger initiated"]:
            if record.levelname in ["INIT", "SUBDEBUG", "DEBUG", "INFO", "ERROR"]:
                assert record.message in ["Mlem",
                                          "{'key': 123}, {'access_token': 'abc'}, {\"access_token\": \"abc\"}",
                                          "https://mlem.api/mlem?access_token=123",
                                          "https://mlem.api/mlem?access_token=123&abs=True"]
            elif record.levelname == "WARNING":
                assert record.message in ["'DarkyLogger' has no attribute 'mlem'. \"INFO\" is used instead.",
                                          "Mlem",
                                          "{'key': 123}, {'access_token': 'abc'}, {\"access_token\": \"abc\"}",
                                          "https://mlem.api/mlem?access_token=123",
                                          "https://mlem.api/mlem?access_token=123&abs=True"]
            elif record.levelname == "CRITICAL":
                assert record.message == "We got an error! Mlem"


from typing import TYPE_CHECKING
import logging
import re
import inspect

from fastapi import APIRouter

if TYPE_CHECKING:
    from ...twilight_vk import TwilightVK

class VkApiRouter:

    def __init__(self, _bot: 'TwilightVK') -> None:
        '''
        Framework's VkApi methods API router

        :param bot: TwilightVK object
        :type bot: TwilightVK
        '''
        self.logger = logging.getLogger("twi-api-vkapi")
        self.logger.log(1, f"VkApi API Router was initialized")
        self.bot = _bot

        self.router = APIRouter(
            tags=[self.bot.bot_name],
            prefix=f"/vk-methods"
        )

        self._parseVkApi()
    
    def _parseVkApi(self) -> None:

        self.logger.debug(f"Getting list of method groups...")
        _methodGroups = self.bot.methods.__dict__
        _methodGroups.__delitem__("logger")
        self.logger.debug(f"List of method groups: {_methodGroups}")

        for key in _methodGroups.keys():
            
            _router = APIRouter(
                    tags=[self.bot.bot_name],
                    prefix=f"/{key}"
                )
            
            self.logger.debug(f"Getting list of the methods inside {key} class...")
            _methodList = inspect.getmembers(getattr(self.bot.methods, key), predicate=inspect.ismethod)
            _methodsList = {_method[0]: _method[1] for _method in _methodList}
            _methodsList.__delitem__("__init__")
            self.logger.debug(f"List of methods: {_methodsList}")

            for _method in _methodsList.keys():
                _router.add_api_route(
                    f"/{_method}", getattr(getattr(self.bot.methods, key), _method), methods=["POST"]
                )

            self.router.include_router(_router)
    
    def get_router(self) -> APIRouter:
        return self.router


from typing import Callable, TYPE_CHECKING
import asyncio

from fastapi import APIRouter

from .base import BaseRouter
from ...utils.types.twi_states import (
    TwiAPIStates,
    TwiVKStates
)

if TYPE_CHECKING:
    from ... import TwilightVK
    from ... import TwilightAPI

class RootRouter(BaseRouter):

    def __init__(self,
                 bots: list['TwilightVK'] = [],
                 api: 'TwilightAPI' = None):
        '''
        Root API router for TwilightAPI

        :param bots: List of bots will be connected to the API
        :type bots: list[TwilightVK]

        :param api: TwilightAPI object
        :type api: TwilightAPI
        '''
        self.router = APIRouter(
            tags=["Root"],
            prefix="/root"
        )

        self._bots = bots
        self._api = api
        
        self.router.add_api_route("/ping", self.ping, methods=["GET"], 
                                  name="Pings the API",
                                  description="Method for API status check. All connected APIs should return \"pong\" if alive.")
        self.router.add_api_route("/stop", self.stop, methods=["GET"],
                                  name="Stops the API",
                                  description="Stops all framework routers of API and main API itself")
    
    async def ping(self) -> dict:
        '''
        API METHOD
        Pings all API routers of bots connected to the TwilightAPI
        '''
        responses = {"server": "Pong OwO"}

        for bot in self._bots:
            responses[f"{bot.bot_name}"] = await bot.api_router.ping()

        return {"response": responses}
    
    async def stop(self, stop_api: bool = False, force: bool = False) -> dict:
        '''
        API METHOD
        Stops all bots connected to the TwilightAPI and their API

        :param stop_api: Should this method also stop the TwilightAPI itself
        :type stop_api: bool

        :param force: Forced stop for TwilightVK bots
        :type force: bool
        '''
        responses = {}

        for bot in self._bots:
            responses[f"{bot.bot_name}"] = await bot.api_router.stop(force)

        while not all([bot._state == TwiVKStates.DISABLED for bot in self._bots]):
            await asyncio.sleep(1.0)

        if self._api and stop_api:
            await self._api.stop()
            responses[f"server"] = "Shutdown initiated"

        return {"response": responses}


import pytest
import pytest_asyncio
import pytest_httpbin
from http import HTTPStatus

from twilight_vk.http.async_http import Http
from twilight_vk.framework.validators.http_validator import HttpValidator
from twilight_vk.framework.validators.event_validator import EventValidator
from twilight_vk.framework.exceptions import (
    HttpValidationError,
    EventValidationError,
    LongPollError,
    VkApiError
)

@pytest_asyncio.fixture
async def http_client():
    httpClient = Http(headers={"X-Test": "test"}, timeout=10)
    yield httpClient
    await httpClient.close()

@pytest.fixture
def httpValidator():
    validator = HttpValidator()
    yield validator

@pytest.fixture
def eventValidator():
    validator = EventValidator()
    yield validator

class MockFakeResponse():

    def __init__(self):
        self.status = HTTPStatus.OK
        self.method = "GET"
        self.url = "https://example.com/"

@pytest.fixture
def fake_events():
    return {
        "polling_success": {
            "ts": 0,
            "updates": []
        },
        "polling_failed": {
            "failed": 3,
        },
        "request_success": {
            "response": {
                "test": 1
            }
        },
        "request_failed": {
            "error": {
                        "error_code": 1,
                        "error_msg": "Unknown error, try again later",
                        "request_params": [
                            {
                                "key": "v",
                                "value": "1.234"
                            }
                        ]
                    }
        }
    }

@pytest.mark.asyncio
async def test_httpValidator(caplog, httpbin, http_client: Http, httpValidator: HttpValidator):
    raw = await http_client.get(
        url = f"{httpbin.url}/get",
        params = {
            "key": "value"
        },
        raw=True
    )
    assert await httpValidator.__isValid__(raw) == True
    assert await httpValidator.__isSuccess__(raw) == True

    invalid_response = MockFakeResponse()
    assert await httpValidator.__isValid__(invalid_response) == False
    assert await httpValidator.__isSuccess__(invalid_response) == True
    
    notsuccess_response = raw
    notsuccess_response.status = HTTPStatus.NOT_FOUND
    assert await httpValidator.__isValid__(notsuccess_response) == True
    assert await httpValidator.__isSuccess__(notsuccess_response) == False

@pytest.mark.asyncio
async def test_eventValidator(caplog, httpbin, http_client: Http, eventValidator: EventValidator, fake_events: dict):
    success_raw = await http_client.post(
        url = f"{httpbin.url}/post",
        data = fake_events.get("polling_success"),
        raw=True
    )
    assert await eventValidator.__isJsonValid__(success_raw) == True
    assert await eventValidator.__fieldsAreValid__((await success_raw.json())["json"], eventValidator.__pollingRequiredFileds__) == True
    assert await eventValidator.__haveErrors__((await success_raw.json())["json"]) == False

    failed_raw = await http_client.post(
        url = f"{httpbin.url}/post",
        data = fake_events.get("polling_failed"),
        raw=True
    )
    assert await eventValidator.__isJsonValid__(failed_raw) == True
    assert await eventValidator.__fieldsAreValid__((await failed_raw.json())["json"], eventValidator.__pollingRequiredFileds__) == True
    with pytest.raises(LongPollError, match=f"3"):
        await eventValidator.__haveErrors__((await failed_raw.json())["json"])

    invalid_raw = await http_client.post(
        url = f"{httpbin.url}/post",
        data = {"key": "value"},
        raw=True
    )
    assert await eventValidator.__isJsonValid__(invalid_raw) == True
    assert await eventValidator.__fieldsAreValid__((await invalid_raw.json())["json"], eventValidator.__pollingRequiredFileds__) == False
    assert await eventValidator.__haveErrors__((await invalid_raw.json())["json"]) == False

    success_raw = await http_client.post(
        url = f"{httpbin.url}/post",
        data = fake_events.get("request_success"),
        raw=True
    )
    assert await eventValidator.__isJsonValid__(success_raw) == True
    assert await eventValidator.__fieldsAreValid__((await success_raw.json())["json"], eventValidator.__requiredFields__) == True
    assert await eventValidator.__haveErrors__((await success_raw.json())["json"]) == False

    failed_raw = await http_client.post(
        url = f"{httpbin.url}/post",
        data = fake_events.get("request_failed"),
        raw=True
    )
    assert await eventValidator.__isJsonValid__(failed_raw) == True
    assert await eventValidator.__fieldsAreValid__((await failed_raw.json())["json"], eventValidator.__requiredFields__) == True
    with pytest.raises(VkApiError, match="1"):
        await eventValidator.__haveErrors__((await failed_raw.json())["json"])

    invalid_raw = await http_client.post(
        url = f"{httpbin.url}/post",
        data = {"key": "value"},
        raw=True
    )
    assert await eventValidator.__isJsonValid__(invalid_raw) == True
    assert await eventValidator.__fieldsAreValid__((await invalid_raw.json())["json"], eventValidator.__requiredFields__) == False
    assert await eventValidator.__haveErrors__((await invalid_raw.json())["json"]) == False


from typing import TYPE_CHECKING
from typing import List, Callable
import logging

import asyncio

from ...utils.config import CONFIG
from ...utils.types.response import ResponseHandler
from ..exceptions.handler import (
    ResponseHandlerError
)
from ..rules import *

if TYPE_CHECKING:
    from ..methods import VkMethods

class BASE_EVENT_HANDLER:

    def __init__(self,
                 vk_methods:'VkMethods',):
        '''
        BASE_EVENT_HANDLER is base handler for vk_api events.
        It contsains the base logic for event handler.
        All child event handlers must inherit it

        :param vk_methods: Initialized VkMethods class which allows to use VK API methods
        :type vk_methods: VkMethods
        '''
        self.logger = logging.getLogger(f"event-handler")
        self.logger.log(1, f"{self.__class__.__name__} event handler is initiated")

        self.vk_methods = vk_methods

        self.__funcs__: List[dict[list[BaseRule], Callable]] = []
    
    def __add__(self, 
                func, 
                rules: list):
        '''
        Allows to add callable functions into this handler
        '''
        self.__funcs__.append(
            {
                "rules": rules,
                "func": func
            }
        )
        self.logger.log(1, f"{func.__name__}() was added to {self.__class__.__name__} "\
                          f"with rules: {[f"{rule.__class__.__name__}" for rule in rules]}")
    
    async def __checkRule__(self,
                            rule: BaseRule, 
                            event: dict):
        '''
        Checking rule result for current function and event
        '''
        if not hasattr(rule, "methods") or rule.methods is None:
            await rule.__linkVkMethods__(self.vk_methods)

        result = await rule.__check__(event)

        return result

    async def __checkRules__(self, func, handler, event):
        '''
        Checks all results
        '''
        self.logger.debug(f"Checking rules for {func.__name__} from {self.__class__.__name__}...")
        rule_results = await asyncio.gather(
            *(self.__checkRule__(rule, event) for rule in handler["rules"]),
            return_exceptions=False
        )
        self.logger.debug(f"Rules check results: {rule_results}")
        self.logger.debug(f"{func.__name__}'s rules was checked")

        should_stop = False

        for rule in rule_results:
            if isinstance(rule, Exception):
                self.logger.error(f"Got an exception in rules check results. [{rule.__class__.__name__}({rule})]")
                should_stop = True
            if rule is False:
                self.logger.debug(f"One of the rules has returned False")
                should_stop = True

        if should_stop:
            return False
        
        return rule_results
    
    async def __extractArgs__(self, rule_results:list):
        '''
        Extracts all args from rule_results after regex rules
        '''
        args = {}
        if isinstance(rule_results, list):
            for result in rule_results:
                if isinstance(result, dict):
                    for key, value in result.items():
                        args.setdefault(key, value)
        return args
   
    async def __handleOutput__(self,
                               func,
                               callback:str|ResponseHandler,
                               event):
        '''
        Handles the output from functions which was added to handler

        :param callback: Function's output
        :type callback: str | ResponseHandler
        '''
        self.logger.debug(f"Handling response for {self.__class__.__name__}.{func.__name__}...")

        if callback is None:
            self.logger.debug(f"Callback is None. Output handling was skipped")
            return True

        if isinstance(callback, str):
            callback = ResponseHandler(
                peer_ids=event["object"]["message"]["peer_id"],
                message=callback,
                forward={
                    "peer_id": event["object"]["message"]["peer_id"],
                    "conversation_message_ids": event["object"]["message"]["conversation_message_id"],
                    "is_reply": True
                }
            )
        if isinstance(callback, ResponseHandler):
            response = await self.vk_methods.messages.send(**callback.getData())
            return response
        raise ResponseHandlerError(callback, isinstance(callback, ResponseHandler | None))
        
    async def __callFunc__(self, handler:dict, event:dict):
        '''
        Executes separate function from self.__funcs__ list
        '''
        func = handler["func"]

        rule_results = await self.__checkRules__(func, handler, event)

        extracted_args = await self.__extractArgs__(rule_results)
        
        if rule_results is not False:
            self.logger.debug(f"Calling {func.__name__} from {self.__class__.__name__}...")
            if asyncio.iscoroutinefunction(func):
                response = await func(event, **extracted_args)
            else:
                response = func(event, **extracted_args)

            result = await self.__handleOutput__(func, response, event)
            if result:
                self.logger.debug(f"{func.__name__} from {self.__class__.__name__} was called")

    async def __execute__(self, event):
        '''
        Handles the event synchronously
        '''
        self.logger.debug(f"{self.__class__.__name__} is working right now...")
        for handler in self.__funcs__:
            await self.__callFunc__(handler, event)
        self.logger.debug(f"{self.__class__.__name__} is finished handling the event")
    
    async def __executeAsync__(self, event):
        '''
        Handles the event asynchronously
        '''
        self.logger.debug(f"{self.__class__.__name__} is working asynchronously right now...")
        await asyncio.gather(
            *(self.__callFunc__(handler, event) for handler in self.__funcs__),
            return_exceptions=False
        )
        self.logger.debug(f"{self.__class__.__name__} is finished asynchronously handling the event")


class DEFAULT_HANDLER(BASE_EVENT_HANDLER):
    pass

class MESSAGE_NEW(BASE_EVENT_HANDLER):
    pass

class MESSAGE_REPLY(BASE_EVENT_HANDLER):
    pass


import pytest
import pytest_asyncio
import pytest_httpbin
from aiohttp import ClientResponse

from twilight_vk.http.async_http import Http

@pytest_asyncio.fixture
async def http_client():
    httpClient = Http(headers={"X-Test": "test"}, timeout=10)
    yield httpClient
    await httpClient.close()

@pytest.mark.asyncio
async def test_init_session(http_client):
    assert http_client.headers == {"X-Test": "test"}
    assert http_client.timeout == 10
    assert http_client.session is None

@pytest.mark.asyncio
async def test_http_get_raw(http_client: Http, httpbin):
    response = await http_client.get(
        url = f"{httpbin.url}/get",
        params = {
            "key": "value"
        },
        raw=True
    )
    assert isinstance(response, ClientResponse)
    assert response.status == 200
    assert (await response.json())["args"] == {"key": "value"}

@pytest.mark.asyncio
async def test_http_get_json(http_client: Http, httpbin):
    response = await http_client.get(
        url = f"{httpbin.url}/get",
        params = {
            "key": "value"
        },
        raw=False
    )
    assert isinstance(response, dict)
    assert response["args"] == {"key": "value"}
    assert response["headers"]["X-Test"] == "test"

@pytest.mark.asyncio
async def test_http_post_raw(http_client: Http, httpbin):
    response = await http_client.post(
        url = f"{httpbin.url}/post",
        data = {
            "key": "value"
        },
        params = {
            "test": "123"
        },
        raw=True
    )
    assert isinstance(response, ClientResponse)
    assert response.status == 200
    assert (await response.json())["json"] == {"key": "value"}
    assert (await response.json())["args"] == {"test": "123"}

@pytest.mark.asyncio
async def test_http_post_json(http_client: Http, httpbin):
    response = await http_client.post(
        url = f"{httpbin.url}/post",
        data = {
            "key": "value"
        },
        params = {
            "test": "123"
        },
        raw=False
    )
    assert isinstance(response, dict)
    assert response["json"] == {"key": "value"}
    assert response["args"] == {"test": "123"}
    assert response["headers"]["X-Test"] == "test"

@pytest.mark.asyncio
async def test_close_session(http_client: Http):
    await http_client.__getSession__()
    assert http_client.session is not None
    assert not http_client.session.closed
    await http_client.close()
    assert http_client.session is None


class FrameworkError(Exception):
    
    def __init__(self,
                 message:str|None=None):
        '''
        Любые ошибки фреймворка

        :param message: Сообщение передаваемое вместе с исключением
        :type message: str | None
        '''
        self.message=message

    def __str__(self):
        return f"Framework got an error! {f"[{self.message}]" if self.message is not None else ""}"
    
class InitializationError(FrameworkError):
    
    def __init__(self,
                 access_token:str,
                 group_id:int):
        '''
        Ошибка инициализации класса TwilightVK

        :param access_token: Токен сообщества для авторизации
        :type access_token: str

        :param group_id: Идентификатор сообщества
        :type access_token: int
        '''
        self.__accessToken__ = access_token
        self.__groupId__ = group_id
    
    def __str__(self):
        return \
        f"{f" : ACCESS_TOKEN is None or empty" if not self.__accessToken__ else ""}"\
        f"{f" : GROUP_ID is None or empty" if not self.__groupId__ else ""}"


import asyncio

from ...http.async_http import Http
from ...logger.darky_logger import DarkyLogger
from ..methods.base import VkBaseMethods

class BaseLongPoll:

    def __init__(self):
        self.httpClient: Http
        self.base_methods: VkBaseMethods
        self.logger: DarkyLogger

    async def auth(self):
        pass

    async def check_event(self):
        pass

    async def listen(self):
        '''
        Listening for events
        '''
        try:
            self.logger.debug(f"Polling was started")
            while not self.__stop__:
                event = await self.check_event()
                
                if "updates" not in event or event["updates"] == []:
                    continue

                yield event
        
        except asyncio.CancelledError:
            self.logger.warning(f"Listening was forcibly canceled (it is not recommend to do this)")
        finally:
            await self.httpClient.close()
            await self.base_methods.close()
            self.logger.debug(f"Polling was stopped")
            self.__stop__ = True
    
    def stop(self, _from="Polling"):
        if self.__stop__ == False:
            if _from == "Polling":
                self.logger.info(f"Polling will be stopped as soon as the current request will be done. Please wait")
            self.__stop__ = True


import logging

from aiohttp import ClientResponse
from fastapi import APIRouter

from ...http.async_http import Http
from ...logger.darky_logger import DarkyLogger
from ...utils.config import CONFIG
from ..validators.http_validator import HttpValidator
from ..validators.event_validator import EventValidator

class VkBaseMethods:

    def __init__(self,
                 url:str,
                 token:str,
                 group:int):
        self.__url__ = url
        self.__token__ = token
        self.__group__ = group
        self.httpValidator = HttpValidator()
        self.eventValidator = EventValidator()
        self.httpClientHeaders = {"Authorization": f"Bearer {token}"}
        self.httpClient = Http(self.httpClientHeaders)
        self.logger = logging.getLogger("vk-methods")

    async def base_get_method(
            self,
            api_method:str,
            values:dict={},
            headers:dict={},
            validate:bool=True
            ) -> ClientResponse:
        valid_values = {}
        for key, value in values.items():
            if value not in ['', None, 'None']:
                valid_values[key] = value
        headers = headers | self.httpClientHeaders
        self.logger.debug(f"Calling HTTP-GET {api_method} method with {valid_values} {headers if headers != {} else ""}...")
        response = await self.httpClient.get(url=f"{self.__url__}/method/{api_method}",
                                            params=valid_values,
                                            headers=headers,
                                            raw=True)
        if validate:
            response = await self.httpValidator.validate(response)
            response = await self.eventValidator.validate(response)

        self.logger.debug(f"Response for {api_method}: {response if isinstance(response, dict) else f"{response.request_info} <{response.status}>"}")
        return response
        
    async def base_post_method(
            self,
            api_method:str,
            values:dict={},
            data:dict={},
            headers:dict={},
            validate:bool=True
            ) -> ClientResponse:
        valid_values = {}
        for key, value in values.items():
            if value not in ['', None]:
                valid_values[key] = value
        headers = headers | self.httpClientHeaders
        self.logger.debug(f"Calling HTTP-POST {api_method} method with {valid_values} {headers if headers != {} else ""}:{data}...")
        response = await self.httpClient.post(url=f"{self.__url__}/method/{api_method}",
                                            params=valid_values,
                                            data=data,
                                            headers=headers,
                                            raw=True)
        if validate:
            response = await self.httpValidator.validate(response)
            response = await self.eventValidator.validate(response)

        self.logger.debug(f"Response for {api_method}: {response if isinstance(response, dict) else f"{response.request_info} <{response.status}>"}")
        return response
    
    async def close(self):
        self.logger.debug("VkBaseMethods was closed")
        await self.httpClient.close()


class BaseMethodsGroup:

    def __init__(self,
                 baseMethods:VkBaseMethods):
        self.__access_token__ = baseMethods.__token__
        self.__group_id__ = baseMethods.__group__
        self.__api_version__ = CONFIG.VK_API.version
        self.base_api = baseMethods
        self.__class_name__ = self.__class__.__name__
        self.logger = logging.getLogger("vk-methods")
        self.logger.log(1, f"Class {self.__class_name__} was initiated")
        self.method = f"{self.__class_name__[0].lower()}{self.__class_name__[1:]}"


import pytest

from twilight_vk.framework.rules import *
from twilight_vk.framework.handlers.event_handlers import DEFAULT_HANDLER

@pytest.fixture()
def fake_event():
    return {
        "group_id": 123,
        "type": "message_new",
        "object": {
            "client_info": {},
            "message": {
                "from_id": 1234,
                "id": 1,
                "fwd_messages": [],
                "attachments": [],
                "conversation_message_id": 1,
                "text": "",
                "peer_id": 2000000001
            }
        }
    }

@pytest.fixture()
def messages_list():
    return {
        "messages": [
            "test",
            "test darky",
            "test test",
            "test [club123|@club123]",
            "[club123|@club123]",
            "[club123|@club123] test",
            "test [id1234|@id1234]",
            "[id1234|@id1234]",
            "[id1234|@id1234] test",
            "darky",
            "test [club123|@club123] [id1234|@id1234] darky"
        ],
        "replies": [
            None,
            {"test": "message"},
            None,
            None,
            {"test": "message"},
            None,
            None,
            {"test": "message"},
            None,
            None,
            {"test": "message"}
        ],
        "fwds": [
            [],
            [],
            [{"test": "message"}],
            [],
            [],
            [{"test": "message"}],
            [],
            [],
            [{"test": "message"}],
            [],
            []
        ],
        "actions": [
            None,
            {"type": "chat_invite_user", "member_id": 1234},
            {"type": "chat_invite_user", "member_id": -123},
            None,
            {"type": "chat_invite_user", "member_id": 1234},
            {"type": "chat_invite_user", "member_id": -123},
            None,
            {"type": "chat_invite_user", "member_id": 123},
            {"type": "chat_invite_user", "member_id": -123},
            None,
            {"type": "chat_invite_user", "member_id": 123}
        ]
    }

@pytest.fixture()
def results():
    return [
        [True, False, {"triggers": ["test"]}, True, False, False, False, False, False, False, True, False, False],
        [True, False, {"triggers": ["test", "darky"]}, True, {"variable": "darky"}, False, False, True, False, False, True, True, False],
        [True, False, {"triggers": ["test"]}, False, {"variable": "test"}, False, False, False, True, False, True, True, True],
        [True, False, {"triggers": ["test"]}, False, {"variable": "[club123|@club123]"}, {"mentions": [{"type": "club", "id": 123, "screen_name": "club123", "text": "@club123"}]}, True, False, False, False, True, False, False],
        [True, False, False, False, False, {"mentions": [{"type": "club", "id": 123, "screen_name": "club123", "text": "@club123"}]}, True, True, False, False, True, True, False],
        [True, False, {"triggers": ["test"]}, False, False, {"mentions": [{"type": "club", "id": 123, "screen_name": "club123", "text": "@club123"}]}, True, False, True, False, True, True, True],
        [True, False, {"triggers": ["test"]}, False, {"variable": "[id1234|@id1234]"}, {"mentions": [{"type": "id", "id": 1234, "screen_name": "id1234", "text": "@id1234"}]}, False, False, False, False, True, False, False],
        [True, False, False, False, False, {"mentions": [{"type": "id", "id": 1234, "screen_name": "id1234", "text": "@id1234"}]}, False, True, False, False, True, True, False],
        [True, False, {"triggers": ["test"]}, False, False, {"mentions": [{"type": "id", "id": 1234, "screen_name": "id1234", "text": "@id1234"}]}, False, False, True, False, True, True, True],
        [True, False, {"triggers": ["darky"]}, False, False, False, False, False, False, False, True, False, False],
        [True, False, {"triggers": ["test", "darky"]}, False, {"variable": "[club123|@club123] [id1234|@id1234] darky"}, {"mentions": [{"type": "club", "id": 123, "screen_name": "club123", "text": "@club123"}, {"type": "id", "id": 1234, "screen_name": "id1234", "text": "@id1234"}]}, True, True, False, False, True, True, False]
    ]

@pytest.fixture
def handler_results():
    return [
        {"triggers": ["test"]},
        {
            "triggers": ["test", "darky"],
            "variable": "darky"
        },
        {
            "triggers": ["test"],
            "variable": "test"
        },
        {
            "triggers": ["test"],
            "variable": "[club123|@club123]",
            "mentions": [{"type": "club", "id": 123, "screen_name": "club123", "text": "@club123"}]
        },
        {"mentions": [{"type": "club", "id": 123, "screen_name": "club123", "text": "@club123"}]},
        {
            "triggers": ["test"],
            "mentions": [{"type": "club", "id": 123, "screen_name": "club123", "text": "@club123"}]
        },
        {
            "triggers": ["test"],
            "variable": "[id1234|@id1234]",
            "mentions": [{"type": "id", "id": 1234, "screen_name": "id1234", "text": "@id1234"}]
        },
        {"mentions": [{"type": "id", "id": 1234, "screen_name": "id1234", "text": "@id1234"}]},
        {
            "triggers": ["test"],
            "mentions": [{"type": "id", "id": 1234, "screen_name": "id1234", "text": "@id1234"}]
        },
        {"triggers": ["darky"]},
        {
            "triggers": ["test", "darky"],
            "variable": "[club123|@club123] [id1234|@id1234] darky",
            "mentions": [{"type": "club", "id": 123, "screen_name": "club123", "text": "@club123"}, 
                         {"type": "id", "id": 1234, "screen_name": "id1234", "text": "@id1234"}]
        }
    ]

class MockVkMethods:

    def __init__(self):
        pass
    
    class messages():
        async def getConversationMembers(peer_id):
            return {
                "response": {
                    "items": [
                        {
                            "member_id": 1234
                        },
                        {
                            "member_id": -123,
                            "is_admin": True
                        }
                    ]
                }
            }
        
@pytest.fixture()
def rules_list():
    rules_lst: list[BaseRule] = [
        TrueRule(),
        FalseRule(),
        ContainsRule(triggers=["test", "darky"], ignore_case=True),
        TextRule(value=["test", "test darky"], ignore_case=True),
        TwiMLRule(value=["test <variable>"], ignore_case=True),
        MentionRule(),
        IsMentionedRule(),
        ReplyRule(),
        ForwardRule(),
        AdminRule(),
        IsAdminRule(),
        InvitedRule(),
        IsInvitedRule()
    ]
    all_rules = []
    for rule in rules_lst:
        rule.methods = MockVkMethods()
        all_rules.append(rule)
    return all_rules

@pytest.mark.asyncio
async def test_rules(fake_event: dict, messages_list: list, results: list, rules_list: list[BaseRule]):
    for test in range(len(messages_list["messages"])):
        fake_event["object"]["message"]["text"] = messages_list["messages"][test]
        fake_event["object"]["message"]["reply_message"] = messages_list["replies"][test]
        if fake_event["object"]["message"]["reply_message"] == None:
            fake_event["object"]["message"].__delitem__("reply_message")
        fake_event["object"]["message"]["fwd_messages"] = messages_list["fwds"][test]
        fake_event["object"]["message"]["action"] = messages_list["actions"][test]
        if fake_event["object"]["message"]["action"] == None:
            fake_event["object"]["message"].__delitem__("action")
        rule_results = [await rule.__check__(fake_event) for rule in rules_list]
        assert rule_results == results[test]

@pytest.mark.asyncio
async def test_handlerinput(fake_event: dict, messages_list: list, results: list, rules_list: list[BaseRule], handler_results: list):
    for test in range(len(messages_list["messages"])):
        fake_event["object"]["message"]["text"] = messages_list["messages"][test]
        rule_results = [await rule.__check__(fake_event) for rule in rules_list]
        handler_result = await DEFAULT_HANDLER(MockVkMethods()).__extractArgs__(rule_results)
        assert handler_result == handler_results[test]